"""Field-level operations and transformations"""

import logging
import re
import warnings
from typing import Any, Callable, Optional, Union

import numpy as np
import pandas as pd
from pandas.api.types import (
    is_float_dtype,
    is_integer_dtype,
    is_numeric_dtype,
    is_object_dtype,
)

from petsard.metadater.datatype import DataType, LogicalType
from petsard.metadater.field.field_types import FieldConfig, FieldMetadata, FieldStats


class TypeMapper:
    """Maps between different type systems"""

    # Pandas to internal type mapping
    PANDAS_TO_METADATER = {
        "bool": DataType.BOOLEAN,
        "int8": DataType.INT8,
        "int16": DataType.INT16,
        "int32": DataType.INT32,
        "int64": DataType.INT64,
        "uint8": DataType.INT16,
        "uint16": DataType.INT32,
        "uint32": DataType.INT64,
        "uint64": DataType.INT64,
        "float16": DataType.FLOAT32,
        "float32": DataType.FLOAT32,
        "float64": DataType.FLOAT64,
        "object": DataType.STRING,
        "string": DataType.STRING,
        "category": DataType.STRING,
        "datetime64[ns]": DataType.TIMESTAMP,
        "datetime64[ns, UTC]": DataType.TIMESTAMP_TZ,
    }

    @staticmethod
    def _safe_dtype(dtype: Any) -> str:
        """Convert various dtype representations to string"""
        if isinstance(dtype, np.dtype):
            return dtype.name
        elif isinstance(dtype, pd.CategoricalDtype):
            return f"category[{dtype.categories.dtype.name}]"
        elif isinstance(dtype, pd.api.extensions.ExtensionDtype):
            return str(dtype)
        elif isinstance(dtype, str):
            return dtype
        elif isinstance(dtype, type):
            return dtype.__name__.lower()
        else:
            return str(dtype)

    @classmethod
    def _pandas_to_metadater(cls, pandas_dtype: str) -> DataType:
        """Convert pandas dtype to Metadater DataType"""
        dtype_str = cls._safe_dtype(pandas_dtype).lower()

        if dtype_str.startswith("datetime64"):
            if "utc" in dtype_str.lower() or "tz" in dtype_str.lower():
                return DataType.TIMESTAMP_TZ
            return DataType.TIMESTAMP

        if dtype_str.startswith("category"):
            return DataType.STRING

        return cls.PANDAS_TO_METADATER.get(dtype_str, DataType.STRING)


class FieldOperations:
    """Operations for field-level metadata and transformations"""

    LOGICAL_CATEGORICAL_TYPE_PERCENTAGE: float = 0.05

    # Logical type detection patterns with confidence thresholds
    LOGICAL_TYPE_PATTERNS: dict[LogicalType, tuple[str, float]] = {
        LogicalType.EMAIL: (r"^[\w\.-]+@[\w\.-]+\.\w+$", 0.95),
        LogicalType.URL: (r"^https?://[^\s]+$", 0.95),
        LogicalType.IP_ADDRESS: (
            r"^(?:(?:25[0-5]|2[0-4][0-9]|[01]?[0-9][0-9]?)\.){3}(?:25[0-5]|2[0-4][0-9]|[01]?[0-9][0-9]?)$",
            0.95,
        ),
        LogicalType.UUID: (
            r"^[0-9a-fA-F]{8}-[0-9a-fA-F]{4}-[0-9a-fA-F]{4}-[0-9a-fA-F]{4}-[0-9a-fA-F]{12}$",
            0.98,
        ),
        # ... 其他 patterns
    }

    def __init__(self):
        self._logger = logging.getLogger(f"PETsARD.{self.__class__.__name__}")
        self.SPECIAL_VALIDATORS = {
            LogicalType.LATITUDE: self._validate_latitude,
            LogicalType.LONGITUDE: self._validate_longitude,
            LogicalType.PERCENTAGE: self._validate_percentage,
            LogicalType.CURRENCY: self._validate_currency,
        }

    @classmethod
    def build_field_from_series(
        cls,
        field_data: pd.Series,
        field_name: str = "Unnamed Field",
        config: Optional[Union[FieldConfig, dict[str, Any]]] = None,
        compute_stats: bool = True,
        infer_logical_type: bool = True,
        optimize_dtype: bool = True,
        sample_size: Optional[int] = 1000,
    ) -> FieldMetadata:
        """Build FieldMetadata from a pandas Series"""
        if field_name == "Unnamed Field" and hasattr(field_data, "name"):
            field_name = str(field_data.name)

        instance = cls()
        instance._logger.debug(
            f"Building field metadata for '{field_name}' with length {len(field_data)}"
        )

        # Convert dict to FieldConfig if needed
        if config is not None and not isinstance(config, FieldConfig):
            if isinstance(config, dict):
                config = FieldConfig(**config)
            else:
                raise TypeError(
                    f"config must be FieldConfig or dict, got {type(config)}"
                )

        if config is None:
            config = FieldConfig()

        # Determine data type
        pandas_dtype: Any = TypeMapper._safe_dtype(field_data.dtype)
        data_type: DataType = TypeMapper._pandas_to_metadater(pandas_dtype)

        # Override with type hint if provided
        if config.type_hint:
            data_type = instance._apply_type_hint(config.type_hint, data_type)

        # Determine nullable
        nullable = (
            config.nullable if config.nullable is not None else field_data.isna().any()
        )

        # Build field metadata
        field_metadata = FieldMetadata(
            name=field_name,
            data_type=data_type,
            nullable=nullable,
            source_dtype=pandas_dtype,
            description=config.description,
            cast_error=config.cast_error,
            properties=config.properties.copy(),
        )

        # Infer logical type
        if infer_logical_type:
            if config.logical_type:
                try:
                    field_metadata.logical_type = LogicalType(config.logical_type)
                except ValueError:
                    instance._logger.warning(
                        f"Invalid logical type '{config.logical_type}' for field '{field_name}'"
                    )
                    field_metadata.logical_type = instance._infer_field_logical_type(
                        field_data=field_data, field_metadata=field_metadata
                    )
            else:
                field_metadata.logical_type = instance._infer_field_logical_type(
                    field_data=field_data, field_metadata=field_metadata
                )

        # Calculate statistics
        if compute_stats:
            sample_data = field_data
            if sample_size and len(field_data) > sample_size:
                sample_data = field_data.sample(n=sample_size, random_state=42)
            field_metadata.stats = instance._calc_field_stats(
                field_data=sample_data, field_metadata=field_metadata
            )

        # Determine optimal target dtype
        if optimize_dtype:
            field_metadata.target_dtype = instance._determine_field_optimal_dtype(
                field_data=field_data, field_metadata=field_metadata
            )

        return field_metadata

    def _apply_type_hint(self, type_hint: str, current_type: DataType) -> DataType:
        """Apply type hint to determine data type"""
        type_hint = type_hint.lower()
        type_mapping = {
            "category": DataType.STRING,
            "datetime": DataType.TIMESTAMP,
            "date": DataType.DATE,
            "time": DataType.TIME,
            "int": DataType.INT64,
            "integer": DataType.INT64,
            "float": DataType.FLOAT64,
            "string": DataType.STRING,
            "boolean": DataType.BOOLEAN,
        }
        return type_mapping.get(type_hint, current_type)

    def _infer_field_logical_type(
        self, field_data: pd.Series, field_metadata: FieldMetadata
    ) -> Optional[LogicalType]:
        """Infer logical type from data patterns"""
        # Only process string-like data
        if field_metadata.data_type not in [DataType.STRING, DataType.BINARY]:
            return None

        sample = field_data.dropna()
        if len(sample) == 0:
            return None

        # Limit sample size
        if len(sample) > 1000:
            sample = sample.sample(n=1000, random_state=42)

        # Ensure string operations are possible
        if not self._is_string_compatible(series=sample):
            return None

        try:
            if not pd.api.types.is_string_dtype(sample):
                sample = sample.astype(str)
        except Exception:
            return None

        # Check patterns
        for logical_type, (pattern, threshold) in self.LOGICAL_TYPE_PATTERNS.items():
            try:
                matches = sample.str.match(pattern, na=False)
                match_ratio = matches.sum() / len(matches)

                if match_ratio >= threshold:
                    if logical_type in self.SPECIAL_VALIDATORS:
                        validator = self.SPECIAL_VALIDATORS[logical_type]
                        if not self._validate_with_function(
                            sample, validator, threshold
                        ):
                            continue
                    return logical_type
            except Exception as e:
                self._logger.debug(f"Pattern matching failed for {logical_type}: {e}")
                continue

        # Check categorical
        unique_ratio = field_data.nunique() / len(field_data)
        if unique_ratio < self.LOGICAL_CATEGORICAL_TYPE_PERCENTAGE:
            return LogicalType.CATEGORICAL

        return None

    def _calc_field_stats(
        self, field_data: pd.Series, field_metadata: FieldMetadata
    ) -> FieldStats:
        """Calculate field statistics"""
        stats = FieldStats(
            row_count=len(field_data),
            na_count=field_data.isna().sum(),
        )
        stats.na_percentage = round((stats.na_count / stats.row_count) * 100, 4)

        # For numeric types
        if field_metadata.data_type in [
            DataType.INT8,
            DataType.INT16,
            DataType.INT32,
            DataType.INT64,
            DataType.FLOAT32,
            DataType.FLOAT64,
        ]:
            if not field_data.dropna().empty:
                stats.min_value = field_data.min()
                stats.max_value = field_data.max()
                stats.mean_value = float(field_data.mean())
                stats.std_value = float(field_data.std())
                stats.quantiles = {
                    0.25: field_data.quantile(0.25),
                    0.5: field_data.quantile(0.5),
                    0.75: field_data.quantile(0.75),
                }

        # For all types
        stats.distinct_count = field_data.nunique()

        # Most frequent values
        value_counts = field_data.value_counts().head(10)
        if not value_counts.empty:
            stats.most_frequent = list(zip(value_counts.index, value_counts.values))

        return stats

    def _determine_field_optimal_dtype(
        self, field_data: pd.Series, field_metadata: FieldMetadata
    ) -> str:
        """Determine optimal dtype for storage"""
        if (
            field_metadata.data_type == DataType.STRING
            and field_metadata.logical_type == LogicalType.CATEGORICAL
        ):
            return "category"

        if is_numeric_dtype(field_data):
            return self._optimize_numeric_dtype(field_data)
        elif is_object_dtype(field_data):
            return self._optimize_object_dtype(field_data)

        return str(field_data.dtype)

    # Helper methods
    def _is_string_compatible(self, series: pd.Series) -> bool:
        """Check if series can use string operations"""
        if pd.api.types.is_string_dtype(series):
            return True
        if pd.api.types.is_object_dtype(series):
            try:
                non_null = series.dropna()
                if len(non_null) == 0:
                    return False
                return all(isinstance(x, str) for x in non_null.head(100))
            except Exception:
                return False
        return False

    def _validate_with_function(
        self, sample: pd.Series, validator: Callable, threshold: float
    ) -> bool:
        """Validate sample using a custom validator function"""
        try:
            valid_count = sum(validator(x) for x in sample)
            return (valid_count / len(sample)) >= threshold
        except Exception:
            return False

    @staticmethod
    def _validate_latitude(value: str) -> bool:
        """Validate latitude"""
        try:
            lat = float(value)
            return -90 <= lat <= 90
        except (ValueError, TypeError):
            return False

    @staticmethod
    def _validate_longitude(value: str) -> bool:
        """Validate longitude"""
        try:
            lon = float(value)
            return -180 <= lon <= 180
        except (ValueError, TypeError):
            return False

    @staticmethod
    def _validate_percentage(value: str) -> bool:
        """Validate percentage"""
        if value.endswith("%"):
            try:
                float(value[:-1])
                return True
            except ValueError:
                return False
        try:
            num = float(value)
            return 0 <= num <= 1 or 0 <= num <= 100
        except ValueError:
            return False

    @staticmethod
    def _validate_currency(value: str) -> bool:
        """Validate currency"""
        currency_patterns = [
            r"^\$[\d,]+\.?\d*$",
            r"^[\d,]+\.?\d*\$$",
            r"^€[\d,]+\.?\d*$",
            r"^£[\d,]+\.?\d*$",
            r"^¥[\d,]+\.?\d*$",
            r"^[A-Z]{3}\s?[\d,]+\.?\d*$",
        ]
        return any(re.match(pattern, value) for pattern in currency_patterns)

    def _optimize_numeric_dtype(self, field_data: pd.Series) -> str:
        """Optimize numeric dtype"""

        if is_integer_dtype(field_data):
            if field_data.isna().all():
                return "int64"

            ranges = {
                "int8": (np.iinfo(np.int8).min, np.iinfo(np.int8).max),
                "int16": (np.iinfo(np.int16).min, np.iinfo(np.int16).max),
                "int32": (np.iinfo(np.int32).min, np.iinfo(np.int32).max),
                "int64": (np.iinfo(np.int64).min, np.iinfo(np.int64).max),
            }
        elif is_float_dtype(field_data):
            if field_data.isna().all():
                return "float32"

            ranges = {
                "float32": (np.finfo(np.float32).min, np.finfo(np.float32).max),
                "float64": (np.finfo(np.float64).min, np.finfo(np.float64).max),
            }
        else:
            return str(field_data.dtype)

        col_min, col_max = np.nanmin(field_data), np.nanmax(field_data)

        for dtype, (min_val, max_val) in ranges.items():
            if min_val <= col_min and col_max <= max_val:
                return dtype

        return str(field_data.dtype)

    def _optimize_object_dtype(self, field_data: pd.Series) -> str:
        """Optimize object dtype"""
        if field_data.isna().all():
            return "category"

        field_data_clean = field_data.dropna()

        # Try datetime conversion
        with warnings.catch_warnings():
            warnings.simplefilter("ignore", UserWarning)
            datetime_field_data = pd.to_datetime(field_data_clean, errors="coerce")

        if not datetime_field_data.isna().any():
            return "datetime64[s]"

        return "category"



"""Metadata layer - Multi-table dataset management"""

from petsard.metadater.metadata.metadata_ops import MetadataOperations
from petsard.metadater.metadata.metadata_types import (
    Metadata,
    MetadataConfig,
    SchemaRelation,
)

__all__ = [
    "Metadata",
    "MetadataConfig",
    "SchemaRelation",
    "MetadataOperations",
]



#!/usr/bin/env python3
"""
Splitter Overlap Control Demo
演示 Splitter 重疊控制功能

這個示例展示了如何使用改進的 Splitter 功能：
1. 控制樣本之間的重疊百分比
2. 追蹤訓練資料的原始索引
3. 函數化編程方式返回結果
"""

import pandas as pd

from petsard.loader.splitter import Splitter


def main():
    print("=== Splitter 重疊控制功能演示 ===\n")

    # 創建示例資料
    data = pd.DataFrame(
        {
            "feature_1": range(50),
            "feature_2": [f"item_{i}" for i in range(50)],
            "target": [i % 3 for i in range(50)],
        }
    )

    print(f"原始資料大小: {len(data)} 行")
    print(f"資料預覽:\n{data.head()}\n")

    # 示例 1: 嚴格重疊控制
    print("=== 示例 1: 嚴格重疊控制 (最大重疊 10%) ===")
    splitter_strict = Splitter(
        num_samples=2,
        train_split_ratio=0.2,
        max_overlap_ratio=0.1,  # 最大重疊 10%
        random_state=42,
    )

    try:
        split_data, metadata, train_indices = splitter_strict.split(data=data)

        print(f"成功生成 {len(split_data)} 個樣本")
        print("訓練索引:")
        for i, indices in enumerate(train_indices):
            indices_list = list(indices)
            print(f"  樣本 {i + 1}: {len(indices_list)} 個索引 - {indices_list[:5]}...")

        # 檢查重疊情況
        print("\n重疊分析:")
        for i in range(len(train_indices)):
            for j in range(i + 1, len(train_indices)):
                overlap = len(train_indices[i].intersection(train_indices[j]))
                overlap_pct = overlap / len(train_indices[i]) * 100
                print(
                    f"  樣本 {i + 1} vs 樣本 {j + 1}: {overlap} 個重疊索引 ({overlap_pct:.1f}%)"
                )

    except Exception as e:
        print(f"嚴格重疊控制失敗: {e}")

    print("\n" + "=" * 60 + "\n")

    # 示例 2: 寬鬆重疊控制
    print("=== 示例 2: 寬鬆重疊控制 (最大重疊 30%) ===")
    splitter_loose = Splitter(
        num_samples=3,
        train_split_ratio=0.3,
        max_overlap_ratio=0.3,  # 最大重疊 30%
        random_state=42,
    )

    split_data, metadata, train_indices = splitter_loose.split(data=data)

    print(f"成功生成 {len(split_data)} 個樣本")
    print("每個樣本的資料大小:")
    for sample_id, sample_data in split_data.items():
        train_size = len(sample_data["train"])
        val_size = len(sample_data["validation"])
        print(f"  樣本 {sample_id}: 訓練集 {train_size} 行, 驗證集 {val_size} 行")

    print("\n重疊分析:")
    for i in range(len(train_indices)):
        for j in range(i + 1, len(train_indices)):
            overlap = len(train_indices[i].intersection(train_indices[j]))
            overlap_pct = overlap / len(train_indices[i]) * 100
            print(
                f"  樣本 {i + 1} vs 樣本 {j + 1}: {overlap} 個重疊索引 ({overlap_pct:.1f}%)"
            )

    print("\n" + "=" * 60 + "\n")

    # 示例 3: 使用 exist_train_indices 功能
    print("=== 示例 3: 使用 exist_train_indices 避免與現有樣本重疊 ===")

    # 假設我們已經有一些現有的樣本索引
    existing_indices = [
        set(range(0, 10)),  # 第一個現有樣本
        set(range(15, 25)),  # 第二個現有樣本
    ]

    splitter_exclude = Splitter(
        num_samples=2,
        train_split_ratio=0.2,
        max_overlap_ratio=0.3,  # 與現有樣本最大重疊 30%
        random_state=42,
    )

    split_data, metadata, train_indices = splitter_exclude.split(
        data=data, exist_train_indices=existing_indices
    )

    print(f"成功生成 {len(split_data)} 個新樣本")
    print("與現有樣本的重疊分析:")

    for i, indices in enumerate(train_indices):
        print(f"\n  新樣本 {i + 1} ({len(indices)} 個索引):")
        for j, existing_set in enumerate(existing_indices):
            overlap = len(indices.intersection(existing_set))
            overlap_pct = overlap / len(indices) * 100
            print(f"    vs 現有樣本 {j + 1}: {overlap} 個重疊索引 ({overlap_pct:.1f}%)")

    print("\n" + "=" * 60 + "\n")

    # 示例 4: 函數化編程 - 不依賴物件狀態
    print("=== 示例 4: 函數化編程方式 ===")

    def create_splits_functional(data, num_samples=2, overlap_pct=0.15):
        """函數化方式創建分割，不依賴物件狀態"""
        splitter = Splitter(
            num_samples=num_samples,
            train_split_ratio=0.3,
            max_overlap_ratio=overlap_pct,
            random_state=42,
        )

        # 直接返回結果，不存儲在物件中
        return splitter.split(data=data)

    # 使用函數化方式
    splits, meta, indices = create_splits_functional(
        data, num_samples=3, overlap_pct=0.2
    )

    print(f"函數化方式成功創建 {len(splits)} 個分割")
    print("返回的結果類型:")
    print(f"  splits: {type(splits)} (包含 {len(splits)} 個樣本)")
    print(f"  metadata: {type(meta)} (包含 {len(meta)} 個樣本的元資料)")
    print(f"  indices: {type(indices)} (包含 {len(indices)} 個索引集合)")

    # 顯示分割資訊
    sample_data = list(splits.values())[0]
    train_size = len(sample_data["train"])
    val_size = len(sample_data["validation"])
    print(f"  分割資訊: 訓練集 {train_size} 行, 驗證集 {val_size} 行")

    print("\n=== 演示完成 ===")


if __name__ == "__main__":
    main()



import unittest
from unittest.mock import patch

import numpy as np
import pandas as pd

from petsard.evaluator.stats import Stats, StatsConfig, StatsJSDivergence, StatsMean


class TestStats(unittest.TestCase):
    """Test for Stats Evaluator."""

    def setUp(self):
        """Set up test fixtures."""
        # Test configuration
        self.config = {
            "eval_method": "stats",
            "stats_method": ["mean", "std", "jsdivergence"],
            "compare_method": "pct_change",
            "aggregated_method": "mean",
            "summary_method": "mean",
        }

        # Create mock data
        self.data = {
            "ori": pd.DataFrame(
                {
                    "col1": [1, 2, 3, 4, 5],
                    "col2": ["a", "b", "c", "d", "e"],
                    "col3": [1.1, 2.2, 3.3, 4.4, 5.5],
                }
            ),
            "syn": pd.DataFrame(
                {
                    "col1": [1, 2, 3, 4, 5],
                    "col2": ["a", "b", "c", "d", "e"],
                    "col3": [1.1, 2.2, 3.3, 4.4, 5.5],
                }
            ),
        }

    def test_init(self):
        """Test initialization."""
        evaluator = Stats(config=self.config)
        self.assertEqual(evaluator.config, self.config)
        self.assertIsNone(evaluator._impl)
        self.assertIsInstance(evaluator.stats_config, StatsConfig)

    @patch.object(StatsMean, "eval")
    @patch.object(StatsJSDivergence, "eval")
    def test_eval_basic(self, mock_js, mock_mean):
        """Test eval method with basic statistics."""
        # Setup mock returns
        mock_mean.return_value = 3.0  # Average of [1, 2, 3, 4, 5]
        mock_js.return_value = 0.0  # JSD is 0 for identical distributions

        # Patch Metadata._convert_dtypes to correctly identify column types
        with patch("petsard.loader.Metadata._convert_dtypes") as mock_convert:
            # Return appropriate dtype identifiers
            def side_effect(dtype):
                if np.issubdtype(dtype, np.number):
                    return "numerical"
                elif dtype == np.dtype("O"):
                    return "categorical"
                return "unknown"

            mock_convert.side_effect = side_effect

            # Execute evaluator
            evaluator = Stats(config=self.config)
            result = evaluator.eval(self.data)

        # Assert results structure
        self.assertIn("global", result)
        self.assertIn("columnwise", result)

        # Assert global results contain a Score
        global_data = result["global"]
        self.assertIsInstance(global_data, pd.DataFrame)
        self.assertIn("Score", global_data.columns)

        # Assert columnwise results contain expected columns
        columnwise_data = result["columnwise"]
        self.assertIsInstance(columnwise_data, pd.DataFrame)

        # Check for specific columns that should exist in results
        expected_cols = [
            "mean_ori",
            "mean_syn",
            "mean_pct_change",
            "std_ori",
            "std_syn",
            "std_pct_change",
        ]
        for col in expected_cols:
            self.assertTrue(
                any(col in columnwise_data.columns for col in expected_cols),
                f"Expected column pattern '{col}' not found",
            )

    def test_eval_with_different_data(self):
        """Test eval with different original and synthetic data."""
        # Create data with different values to test comparison
        different_data = {
            "ori": pd.DataFrame(
                {
                    "col1": [1, 2, 3, 4, 5],
                    "col3": [1.0, 2.0, 3.0, 4.0, 5.0],
                }
            ),
            "syn": pd.DataFrame(
                {
                    "col1": [2, 3, 4, 5, 6],  # Shifted by 1
                    "col3": [2.0, 4.0, 6.0, 8.0, 10.0],  # Doubled
                }
            ),
        }

        # Patch Metadata._convert_dtypes to correctly identify column types
        with patch("petsard.loader.Metadata._convert_dtypes") as mock_convert:
            mock_convert.return_value = (
                "numerical"  # Simplify by making all columns numerical
            )

            # Test with both comparison methods
            for compare_method in ["diff", "pct_change"]:
                config = self.config.copy()
                config["compare_method"] = compare_method
                config["stats_method"] = ["mean"]  # Simplify test to just mean

                evaluator = Stats(config=config)
                result = evaluator.eval(different_data)

                # Verify results contain the comparison column
                self.assertIn("columnwise", result)
                columnwise = result["columnwise"]
                comparison_col = f"mean_{compare_method}"
                self.assertIn(comparison_col, columnwise.columns)

                # Basic check that values are different from 0 (since data is different)
                self.assertTrue((columnwise[comparison_col].abs() > 0).any())

    def test_invalid_stats_method(self):
        """Test with invalid stats method."""
        invalid_config = self.config.copy()
        invalid_config["stats_method"] = ["invalid_method"]

        with self.assertRaises(Exception):
            Stats(config=invalid_config)

    def test_invalid_compare_method(self):
        """Test with invalid comparison method."""
        invalid_config = self.config.copy()
        invalid_config["compare_method"] = "invalid_method"

        with self.assertRaises(Exception):
            Stats(config=invalid_config)


if __name__ == "__main__":
    unittest.main()



import logging
import os
import time
from dataclasses import dataclass
from datetime import datetime, timedelta

import yaml

from petsard.config import Config
from petsard.config_base import BaseConfig
from petsard.exceptions import ConfigError
from petsard.status import Status


@dataclass
class ExecutorConfig(BaseConfig):
    """
    Defines the configuration for the Executor.

    Attr.:
        _logger (logging.Logger): The logger object.
        log_output_type (str): Output destination
            - stdout
            - file
            - both
        log_level (str): Logging level
            - DEBUG
            - INFO
            - WARNING
            - ERROR
            - CRITICAL
        log_dir (str): Directory for storing log files
        log_name (str): Log file name template (can include {timestamp})
    """

    log_output_type: str = "file"
    log_level: str = "INFO"
    log_dir: str = "."
    log_filename: str = "PETsARD_{timestamp}.log"

    def __post_init__(self):
        """
        Post-initialization method to validate the configuration.
        """
        super().__post_init__()
        if self.log_output_type not in ["stdout", "file", "both"]:
            raise ConfigError("Invalid log_output_type {self.log_output_type}")
        if self.log_level not in ["DEBUG", "INFO", "WARNING", "ERROR", "CRITICAL"]:
            raise ConfigError("Invalid log_level {self.log_level}")


class Executor:
    """
    Represents an executor that runs a series of operators based on a given configuration.
    """

    def __init__(self, config: str):
        """
        Args:
            config (str): The configuration filename for the executor.

        Attributes:
            executor_config (ExecutorConfig): The configuration for the executor.
            _logger (logging.Logger): The logger object.
            config (Config): The configuration object.
            status (Status): The status of the executor.
            result (dict): The result of the executor.
        """

        # 1. set the default Executor
        self.executor_config = ExecutorConfig()

        # 2. set the default logger
        self._setup_logger()

        # 3. load the configuration
        self._logger.info(f"Loading configuration from {config}")
        yaml_config: dict = self._get_config(yaml_file=config)

        self.config = Config(config=yaml_config)
        self.sequence = self.config.sequence
        self.status = Status(config=self.config)
        self.result: dict = {}

    def _setup_logger(self, reconfigure=False):
        """
        Setting up the logger based on ExecutorConfig settings.

        Args:
            reconfigure (bool): If True, clear existing handlers before setup
        """
        if reconfigure:
            # If reconfigure, clear existing handlers
            root_logger = logging.getLogger("PETsARD")
            for handler in root_logger.handlers[:]:
                root_logger.removeHandler(handler)
        else:
            # If new configuration, disable the root logger
            logging.getLogger().handlers = []
            root_logger = logging.getLogger("PETsARD")

        # setup logging level
        root_logger.setLevel(getattr(logging, self.executor_config.log_level.upper()))

        # setup formatter
        formatter = logging.Formatter(
            "%(asctime)s - "  # timestamp
            "%(name)-21s - "  # logger name (left align w/ 21 digits: 'PETsARD.Postprocessor')
            "%(funcName)-17s - "  # function name (left align w/ 17 digits: 'inverse_transform')
            "%(levelname)-8s - "  # logger level (left align w/ 8 digits: 'CRITICAL')
            "%(message)s"  # message
        )

        # Handle file output
        if self.executor_config.log_output_type in ["file", "both"]:
            log_dir = self.executor_config.log_dir

            # Create log directory if it doesn't exist
            if log_dir and not os.path.exists(log_dir):
                os.makedirs(log_dir)

            # Create log file
            timestamp = datetime.now().strftime("%Y-%m-%d_%H-%M-%S")
            log_filename = self.executor_config.log_filename.replace(
                "{timestamp}", timestamp
            )
            log_path = os.path.join(log_dir, log_filename)

            # Create file handler
            file_handler = logging.FileHandler(log_path)
            file_handler.setFormatter(formatter)
            root_logger.addHandler(file_handler)

        # Handle stdout output
        if self.executor_config.log_output_type in ["stdout", "both"]:
            console_handler = logging.StreamHandler()
            console_handler.setFormatter(formatter)
            root_logger.addHandler(console_handler)

        # setup this logger as a child of root logger
        self._logger = logging.getLogger(f"PETsARD.{self.__class__.__name__}")

    def _get_config(self, yaml_file: str) -> dict:
        """
        Load the configuration from a YAML file.
        """
        if not os.path.isfile(yaml_file):
            raise ConfigError(f"YAML file {yaml_file} does not exist")

        yaml_config: dict = {}
        with open(yaml_file) as yaml_file:
            yaml_config = yaml.safe_load(yaml_file)

        if "Executor" in yaml_config:
            self.executor_config.update(yaml_config["Executor"])

            self._setup_logger(reconfigure=True)
            self._logger.info("Logger reconfigured with settings from YAML")

            yaml_config.pop("Executor")

        return yaml_config

    def run(self):
        """
        run(): Runs the operators based on the configuration.
        """
        start_time: time = time.time()
        self._logger.info("Starting PETsARD execution workflow")
        while self.config.config.qsize() > 0:
            ops = self.config.config.get()
            module = self.config.module_flow.get()
            expt = self.config.expt_flow.get()

            self._logger.info(f"Executing {module} with {expt}")
            ops.run(ops.set_input(status=self.status))

            self.status.put(module, expt, ops)

            # collect result
            self._set_result(module)

        elapsed_time: time = time.time() - start_time
        formatted_elapsed_time: str = str(timedelta(seconds=round(elapsed_time)))
        self._logger.info(
            f"Completed PETsARD execution workflow (elapsed: {formatted_elapsed_time})"
        )

    def _set_result(self, module: str):
        """
        Get the result for a final module.

        Args:
            module (str): The name of the module.

        Returns:
            None. Update in self.result
        """
        if module == self.sequence[-1]:
            self._logger.debug(f"Collecting final results for {module}")
            full_expt = self.status.get_full_expt()
            full_expt_name = "_".join(
                [f"{module}[{expt}]" for module, expt in full_expt.items()]
            )
            self.result[full_expt_name] = self.status.get_result(module=module)

    def get_result(self):
        """
        Returns the result of the executor.
        """
        return self.result

    def get_timing(self):
        """
        取得執行時間記錄資料

        Returns:
            pd.DataFrame: 包含所有模組執行時間的 DataFrame
        """
        return self.status.get_timing_report_data()



import numpy as np
import pandas as pd
import pytest

from petsard.exceptions import UnfittedError
from petsard.processor.missing import (
    MissingDrop,
    MissingMean,
    MissingMedian,
    MissingSimple,
)


class Test_MissingMean:
    def test_mean_no_missing_values(self):
        # Prepare test data
        df_data = pd.DataFrame({"col1": [1.0, 2.0, 3.0]})

        # Create an instance of the class
        missing = MissingMean()
        missing.set_na_percentage(0.8)

        with pytest.raises(ValueError):
            missing.set_na_percentage(1.8)

        with pytest.raises(ValueError):
            missing.set_na_percentage(-1.8)

        with pytest.raises(UnfittedError):
            missing.transform(df_data["col1"])

        # Call the method to be tested
        missing.fit(df_data["col1"])

        transformed = missing.transform(df_data["col1"])

        rtransform = missing.inverse_transform(df_data["col1"])

        # Assert the result
        assert (transformed.values == np.array([1.0, 2.0, 3.0])).all()

        # Assert the result
        assert transformed.shape == (3,)
        assert rtransform.isna().any().any()

    def test_mean_with_missing_values(self):
        # Prepare test data
        df_data = pd.DataFrame({"col1": [1.0, None, 3.0]})
        df_expected = pd.Series(data=[1.0, 2.0, 3.0], name="col1")

        # Create an instance of the class
        missing = MissingMean()
        missing.set_na_percentage(0.8)

        with pytest.raises(ValueError):
            missing.set_na_percentage(1.8)

        with pytest.raises(ValueError):
            missing.set_na_percentage(-1.8)

        with pytest.raises(UnfittedError):
            missing.transform(df_data["col1"])

        # Call the method to be tested
        missing.fit(df_data["col1"])

        transformed = missing.transform(df_data["col1"])

        rtransform = missing.inverse_transform(df_data["col1"])

        # Assert the result
        assert (transformed.values == np.array([1.0, 2.0, 3.0])).all()

        # Assert the result
        assert transformed.shape == (3,)
        assert transformed.equals(df_expected)
        assert rtransform.isna().any().any()


class Test_MissingMedian:
    def test_median_no_missing_values(self):
        # Prepare test data
        df_data = pd.DataFrame({"col1": [1.0, 2.0, 3.0]})

        # Create an instance of the class
        missing = MissingMedian()
        missing.set_na_percentage(0.8)

        with pytest.raises(ValueError):
            missing.set_na_percentage(1.8)

        with pytest.raises(ValueError):
            missing.set_na_percentage(-1.8)

        with pytest.raises(UnfittedError):
            missing.transform(df_data["col1"])

        # Call the method to be tested
        missing.fit(df_data["col1"])

        transformed = missing.transform(df_data["col1"])

        rtransform = missing.inverse_transform(df_data["col1"])

        # Assert the result
        assert (transformed.values == np.array([1.0, 2.0, 3.0])).all()

        # Assert the result
        assert transformed.shape == (3,)
        assert rtransform.isna().any().any()

    def test_median_with_missing_values(self):
        # Prepare test data
        df_data = pd.DataFrame({"col1": [1.0, None, 3.0]})
        df_expected = pd.Series(data=[1.0, 2.0, 3.0], name="col1")

        # Create an instance of the class
        missing = MissingMedian()
        missing.set_na_percentage(0.8)

        with pytest.raises(ValueError):
            missing.set_na_percentage(1.8)

        with pytest.raises(ValueError):
            missing.set_na_percentage(-1.8)

        with pytest.raises(UnfittedError):
            missing.transform(df_data["col1"])

        # Call the method to be tested
        missing.fit(df_data["col1"])

        transformed = missing.transform(df_data["col1"])

        rtransform = missing.inverse_transform(df_data["col1"])

        # Assert the result
        assert (transformed.values == np.array([1.0, 2.0, 3.0])).all()

        # Assert the result
        assert transformed.shape == (3,)
        assert transformed.equals(df_expected)
        assert rtransform.isna().any().any()


class Test_MissingSimple:
    def test_simple_no_missing_values(self):
        # Prepare test data
        df_data = pd.DataFrame({"col1": [1.0, 2.0, 3.0]})

        # Create an instance of the class
        missing = MissingSimple(value=1.0)
        missing.set_na_percentage(0.8)

        with pytest.raises(ValueError):
            missing.set_na_percentage(1.8)

        with pytest.raises(ValueError):
            missing.set_na_percentage(-1.8)

        with pytest.raises(UnfittedError):
            missing.transform(df_data["col1"])

        # Call the method to be tested
        missing.fit(df_data["col1"])

        transformed = missing.transform(df_data["col1"])

        rtransform = missing.inverse_transform(df_data["col1"])

        # Assert the result
        assert (transformed.values == np.array([1.0, 2.0, 3.0])).all()

        # Assert the result
        assert transformed.shape == (3,)
        assert rtransform.isna().any().any()

    def test_simple_with_missing_values(self):
        # Prepare test data
        df_data = pd.DataFrame({"col1": [1.0, None, 3.0]})
        df_expected = pd.Series(data=[1.0, 2.0, 3.0], name="col1")

        # Create an instance of the class
        missing = MissingSimple(value=2.0)
        missing.set_na_percentage(0.8)

        with pytest.raises(ValueError):
            missing.set_na_percentage(1.8)

        with pytest.raises(ValueError):
            missing.set_na_percentage(-1.8)

        with pytest.raises(UnfittedError):
            missing.transform(df_data["col1"])

        # Call the method to be tested
        missing.fit(df_data["col1"])

        transformed = missing.transform(df_data["col1"])

        rtransform = missing.inverse_transform(df_data["col1"])

        # Assert the result
        assert (transformed.values == np.array([1.0, 2.0, 3.0])).all()

        # Assert the result
        assert transformed.shape == (3,)
        assert transformed.equals(df_expected)
        assert rtransform.isna().any().any()


class Test_MissingDrop:
    def test_drop_no_missing_values(self):
        # Prepare test data
        df_data = pd.DataFrame({"col1": [1.0, 2.0, 3.0]})

        # Create an instance of the class
        missing = MissingDrop()
        missing.set_na_percentage(0.8)

        with pytest.raises(ValueError):
            missing.set_na_percentage(1.8)

        with pytest.raises(ValueError):
            missing.set_na_percentage(-1.8)

        with pytest.raises(UnfittedError):
            missing.transform(df_data["col1"])

        # Call the method to be tested
        missing.fit(df_data["col1"])

        transformed = missing.transform(df_data["col1"])

        rtransform = missing.inverse_transform(df_data["col1"])

        # Assert the result
        assert (transformed == np.array([False, False, False])).all()

        # Assert the result
        assert transformed.shape == (3,)
        assert rtransform.isna().any().any()

    def test_drop_with_missing_values(self):
        # Prepare test data
        df_data = pd.DataFrame({"col1": [1.0, None, 3.0]})

        # Create an instance of the class
        missing = MissingDrop()
        missing.set_na_percentage(0.8)

        with pytest.raises(ValueError):
            missing.set_na_percentage(1.8)

        with pytest.raises(ValueError):
            missing.set_na_percentage(-1.8)

        with pytest.raises(UnfittedError):
            missing.transform(df_data["col1"])

        # Call the method to be tested
        missing.fit(df_data["col1"])

        transformed = missing.transform(df_data["col1"])

        rtransform = missing.inverse_transform(df_data["col1"])

        # Assert the result
        assert (transformed == np.array([False, True, False])).all()

        # Assert the result
        assert transformed.shape == (3,)
        assert rtransform.isna().any().any()



"""Tests for field functions module
測試欄位函數模組
"""

import numpy as np
import pandas as pd

from petsard.metadater.field.field_functions import (
    _comprehensive_type_analysis,
    _has_leading_zeros,
    build_field_metadata,
)
from petsard.metadater.field.field_types import FieldConfig


class TestComprehensiveTypeAnalysis:
    """Test cases for comprehensive type analysis
    完整類型分析的測試案例
    """

    def test_leading_zero_detection(self):
        """Test leading zero detection and preservation
        測試前導零檢測和保留
        """
        # Test data with leading zeros
        leading_zero_data = pd.Series(["001", "002", "010", "099"])
        config = FieldConfig(auto_detect_leading_zeros=True)

        result = _comprehensive_type_analysis(leading_zero_data, config)
        assert result == "string", f"前導零應該返回 string，但得到 {result}"

    def test_float_detection(self):
        """Test float number detection
        測試浮點數檢測
        """
        float_data = pd.Series(["1.5", "2.7", "3.14", "4.0"])
        config = FieldConfig()

        result = _comprehensive_type_analysis(float_data, config)
        assert result in ["float32", "float64"], (
            f"浮點數應該返回 float，但得到 {result}"
        )

    def test_integer_with_nulls(self):
        """Test integer with null values using nullable integers
        測試含空值整數使用 nullable integer
        """
        int_with_na = pd.Series(["10", "20", np.nan, "30"])
        config = FieldConfig(force_nullable_integers=True)

        result = _comprehensive_type_analysis(int_with_na, config)
        assert result.startswith("Int"), f"含空值整數應該返回 Int，但得到 {result}"

    def test_integer_without_nulls(self):
        """Test integer without null values
        測試無空值整數
        """
        int_no_na = pd.Series(["10", "20", "30", "40"])
        config = FieldConfig()

        result = _comprehensive_type_analysis(int_no_na, config)
        assert result.startswith("int"), f"無空值整數應該返回 int，但得到 {result}"

    def test_mixed_non_numeric_data(self):
        """Test mixed non-numeric data
        測試混合非數值資料
        """
        mixed_data = pd.Series(["abc", "def", "ghi"])
        config = FieldConfig()

        result = _comprehensive_type_analysis(mixed_data, config)
        assert result == "category", f"混合資料應該返回 category，但得到 {result}"

    def test_numeric_conversion_threshold(self):
        """Test numeric conversion threshold (80%)
        測試數值轉換門檻值 (80%)
        """
        # 60% numeric data (below threshold)
        mixed_data = pd.Series(["10", "20", "abc", "def", "ghi"])
        config = FieldConfig()

        result = _comprehensive_type_analysis(mixed_data, config)
        assert result == "category", (
            f"低於門檻值的混合資料應該返回 category，但得到 {result}"
        )

    def test_integer_dtype_handling(self):
        """Test handling of different integer dtypes from pd.to_numeric
        測試處理 pd.to_numeric 產生的不同整數類型
        """
        # This will be converted to int64 by pd.to_numeric
        int_data = pd.Series(["100", "200", "300"])
        config = FieldConfig()

        result = _comprehensive_type_analysis(int_data, config)
        assert result.startswith("int"), f"整數資料應該返回 int 類型，但得到 {result}"


class TestLeadingZeroDetection:
    """Test cases for leading zero detection
    前導零檢測的測試案例
    """

    def test_has_leading_zeros_positive(self):
        """Test positive cases for leading zero detection
        測試前導零檢測的正面案例
        """
        # More than 30% have leading zeros
        data = pd.Series(["001", "002", "010", "099", "100"])  # 4/5 = 80%
        assert _has_leading_zeros(data) is True

    def test_has_leading_zeros_negative(self):
        """Test negative cases for leading zero detection
        測試前導零檢測的負面案例
        """
        # Less than 30% have leading zeros
        data = pd.Series(["100", "200", "300", "001"])  # 1/4 = 25% < 30%
        assert _has_leading_zeros(data) is False

    def test_has_leading_zeros_empty_data(self):
        """Test leading zero detection with empty data
        測試空資料的前導零檢測
        """
        data = pd.Series([], dtype=object)
        assert _has_leading_zeros(data) is False

    def test_has_leading_zeros_all_na(self):
        """Test leading zero detection with all NA values
        測試全為 NA 值的前導零檢測
        """
        data = pd.Series([np.nan, np.nan, np.nan])
        assert _has_leading_zeros(data) is False

    def test_has_leading_zeros_mixed_types(self):
        """Test leading zero detection with mixed data types
        測試混合資料類型的前導零檢測
        """
        data = pd.Series(["001", "002", "abc", "010"])  # 3/4 = 75%
        assert _has_leading_zeros(data) is True


class TestFieldMetadataIntegration:
    """Test cases for field metadata integration with new features
    欄位元資料與新功能整合的測試案例
    """

    def test_build_field_metadata_with_leading_zeros(self):
        """Test building field metadata with leading zero detection
        測試建立含前導零檢測的欄位元資料
        """
        data = pd.Series(["001", "002", "003"])

        # With leading zero detection enabled
        config_on = FieldConfig(auto_detect_leading_zeros=True)
        metadata_on = build_field_metadata(data, "test_field", config_on)
        assert metadata_on.target_dtype == "string"

        # With leading zero detection disabled
        config_off = FieldConfig(auto_detect_leading_zeros=False)
        metadata_off = build_field_metadata(data, "test_field", config_off)
        # Should be treated as numeric and converted to int
        assert metadata_off.target_dtype.startswith("int")

    def test_build_field_metadata_with_nullable_integers(self):
        """Test building field metadata with nullable integers
        測試建立含 nullable integer 的欄位元資料
        """
        data = pd.Series([1, 2, np.nan, 4], dtype="float64")

        # With nullable integers enabled
        config_on = FieldConfig(force_nullable_integers=True)
        metadata_on = build_field_metadata(data, "test_field", config_on)
        assert metadata_on.target_dtype.startswith("Int")

        # With nullable integers disabled
        config_off = FieldConfig(force_nullable_integers=False)
        metadata_off = build_field_metadata(data, "test_field", config_off)
        # Should remain as regular int (but might be optimized)
        assert metadata_off.target_dtype.startswith("int")

    def test_build_field_metadata_dtype_optimization(self):
        """Test dtype optimization in field metadata
        測試欄位元資料中的類型優化
        """
        # Small integers should be optimized to int8
        small_int_data = pd.Series([1, 2, 3, 4])
        metadata = build_field_metadata(small_int_data, "small_int")
        assert metadata.target_dtype == "int8"

        # Large integers should use int64
        large_int_data = pd.Series([1000000, 2000000, 3000000])
        metadata = build_field_metadata(large_int_data, "large_int")
        assert metadata.target_dtype in ["int32", "int64"]

        # Small floats should be optimized to float32
        small_float_data = pd.Series([1.1, 2.2, 3.3])
        metadata = build_field_metadata(small_float_data, "small_float")
        assert metadata.target_dtype == "float32"


class TestAmbiguousDataScenarios:
    """Test cases for ambiguous data type scenarios
    容易誤判、型別判斷模糊資料情境的測試案例
    """

    def test_id_code_preservation(self):
        """Test preservation of ID codes with leading zeros
        測試保留前導零的識別代號
        """
        id_codes = pd.Series(["001", "002", "010", "099", "100"])
        config = FieldConfig(auto_detect_leading_zeros=True)

        result = _comprehensive_type_analysis(id_codes, config)
        assert result == "string", "識別代號應該保留為字串類型"

    def test_demographic_data_with_missing_values(self):
        """Test demographic data with missing values
        測試含缺失值的人口統計資料
        """
        demographic_data = pd.Series([25, 30, np.nan, 45, 50], dtype="float64")
        config = FieldConfig(force_nullable_integers=True)

        metadata = build_field_metadata(demographic_data, "demographic", config)
        assert metadata.target_dtype.startswith("Int"), (
            "人口統計資料應該使用 nullable integer"
        )

    def test_financial_amount_detection(self):
        """Test financial amount data as float numbers
        測試金額資料的浮點數檢測
        """
        amount_data = pd.Series([50000.5, 60000.7, np.nan, 75000.3, 80000.9])
        config = FieldConfig()

        result = _comprehensive_type_analysis(amount_data, config)
        assert result.startswith("float"), "金額資料應該被識別為浮點數"

    def test_score_integer_detection(self):
        """Test score data as integers
        測試分數資料的整數檢測
        """
        score_data = pd.Series([85, 90, 78, 92, 88])
        config = FieldConfig()

        metadata = build_field_metadata(score_data, "score", config)
        assert metadata.target_dtype.startswith("int"), "分數資料應該被識別為整數"

    def test_categorical_data_detection(self):
        """Test categorical data detection
        測試分類資料檢測
        """
        category_data = pd.Series(["A級", "B級", "C級", "A級", "B級"])
        config = FieldConfig()

        metadata = build_field_metadata(category_data, "category", config)
        assert metadata.target_dtype == "category", "分類資料應該被識別為分類資料"


class TestEdgeCases:
    """Test cases for edge cases and error handling
    邊界情況和錯誤處理的測試案例
    """

    def test_empty_series(self):
        """Test handling of empty series
        測試處理空序列
        """
        empty_data = pd.Series([], dtype=object)
        config = FieldConfig()

        result = _comprehensive_type_analysis(empty_data, config)
        assert result == "category", "空序列應該返回 category"

    def test_all_null_series(self):
        """Test handling of all-null series
        測試處理全空值序列
        """
        null_data = pd.Series([np.nan, np.nan, np.nan])
        config = FieldConfig()

        result = _comprehensive_type_analysis(null_data, config)
        # 全空值序列會被 pd.to_numeric 轉換為 float，然後優化為 float32
        assert result in ["float32", "category"], (
            "全空值序列應該返回 float32 或 category"
        )

    def test_single_value_series(self):
        """Test handling of single value series
        測試處理單值序列
        """
        single_data = pd.Series(["001"])
        config = FieldConfig(auto_detect_leading_zeros=True)

        result = _comprehensive_type_analysis(single_data, config)
        assert result == "string", "單個前導零值應該被識別為字串"

    def test_mixed_numeric_string_data(self):
        """Test handling of mixed numeric and string data
        測試處理混合數值和字串資料
        """
        mixed_data = pd.Series(["10", "20", "abc", "30", "def"])
        config = FieldConfig()

        result = _comprehensive_type_analysis(mixed_data, config)
        assert result == "category", "混合數值字串資料應該返回 category"

    def test_config_none_handling(self):
        """Test handling when config is None
        測試 config 為 None 的處理
        """
        data = pd.Series([1, 2, 3])

        result = _comprehensive_type_analysis(data, None)
        assert result.startswith("int"), "None config 應該使用預設配置"



import random
from typing import Optional, Union

import pandas as pd

from petsard.exceptions import ConfigError
from petsard.loader.loader import Loader
from petsard.metadater import SchemaMetadata


class Splitter:
    """
    Splitter is an independent module for Executor use. Included:
    a.) split input data via assigned ratio (train_split_ratio)
    b.) resampling assigned times (num_samples)
    c.) output their train/validation indexes (self.index_samples) and pd.DataFrame data (self.data)

    When method is 'custom_data', the data will be loaded from the filepath.
    """

    def __init__(
        self,
        method: str = None,
        num_samples: Optional[int] = 1,
        train_split_ratio: Optional[float] = 0.8,
        random_state: Optional[Union[int, float, str]] = None,
        max_overlap_ratio: Optional[float] = 1.0,
        max_attempts: Optional[int] = 30,
        **kwargs,
    ):
        """
        Args:
            method (str, optional):
                Supports loading existing split data, only accepting 'custom_data'.
                Default is None.
            num_samples (int, optional):
                Number of times to resample the data. Default is 1.
            train_split_ratio (float, optional):
                Ratio of data to assign to the training set,
                must between 0 ~ 1. Default is 0.8.
            random_state (int | float | str, optional):
                Seed for random number generation. Default is None.
            max_overlap_ratio (float, optional):
                Maximum allowed overlap ratio between samples.
                Default is 1.0 (100%). Set to 0.0 for no overlap.
            max_attempts (int, optional):
                Maximum number of attempts for sampling. Default is 30.
            **kwargs (optional):
                For method 'custom_data' only. Apply Loader's config.

        Attr:
            config (dict):
                The configuration of Splitter.
                If method is None,
                    it contains num_samples, train_split_ratio, random_state, max_overlap_ratio, max_attempts.
                If method is 'custom_data',
                    it contains method, filepath, and Loader's config.

        """
        self.config: dict = {}

        # Normal Splitter use case
        if method is None:
            if not (0 <= train_split_ratio <= 1):
                raise ConfigError(
                    "Splitter:  train_split_ratio must be a float between 0 and 1."
                )
            if not (0 <= max_overlap_ratio <= 1):
                raise ConfigError(
                    "Splitter: max_overlap_ratio must be a float between 0 and 1."
                )
            self.config = {
                "num_samples": num_samples,
                "train_split_ratio": train_split_ratio,
                "random_state": random_state,
                "max_overlap_ratio": max_overlap_ratio,
                "max_attempts": max_attempts,
            }

        # custom_data Splitter use case
        else:
            if method.lower() != "custom_data":
                raise ConfigError

            filepath = kwargs.get("filepath", None)
            if filepath is None or not isinstance(filepath, dict):
                raise ConfigError
            if not all(k in filepath for k in ("ori", "control")):
                raise ConfigError

            config = kwargs
            self.loader: dict = {}

            for key in ["ori", "control"]:
                self.loader[key] = Loader(
                    filepath=filepath[key],
                    **{
                        k: config.get(k)
                        for k in [
                            "column_types",
                            "header_names",
                            "na_values",
                        ]
                        if config.get(k) is not None
                    },
                )

            config["method"] = method
            config["filepath"] = filepath
            self.config = config

    def split(
        self,
        data: pd.DataFrame = None,
        exist_train_indices: list[set] = None,
    ) -> tuple[dict, dict, list[set]]:
        """
        Perform index bootstrapping on the Splitter-initialized data
            and split it into train and validation sets
            using the generated index samples.

        When method is 'custom_data', the data will be loaded from the filepath.

        Args:
            data (pd.DataFrame, optional): The dataset which wait for split.
            exist_train_indices (list[set], optional):
                The existing train index sets we want to avoid overlapping with.

        Returns:
            tuple[dict, dict, list[set]]:
                - Split data: {1: {train: pd.DataFrame, validation: pd.DataFrame}, 2: ...}
                - Metadata: {1: {train: SchemaMetadata, validation: SchemaMetadata}, 2: ...}
                - Train indices: [{train_indices_set1}, {train_indices_set2}, ...]
        """
        if "method" in self.config:
            # Custom data method - load from files
            ori_data, ori_metadata = self.loader["ori"].load()
            ctrl_data, _ = self.loader["control"].load()

            split_data = {
                1: {
                    "train": ori_data,
                    "validation": ctrl_data,
                }
            }

            # Create metadata for both train and validation
            train_metadata = self._update_metadata_with_split_info(
                ori_metadata, ori_data.shape[0], ctrl_data.shape[0]
            )
            validation_metadata = self._create_split_metadata(
                ori_data.shape[0], ctrl_data.shape[0]
            )

            metadata_dict = {
                1: {
                    "train": train_metadata,
                    "validation": validation_metadata,
                }
            }

            train_indices_list = [set(ori_data.index.tolist())]

        else:
            # Normal splitting method
            if data is None:
                raise ConfigError("Data must be provided for normal splitting method")

            data.reset_index(drop=True, inplace=True)  # avoid unexpected index

            index_result = self._bootstrapping(
                index=data.index.tolist(), exist_train_indices=exist_train_indices
            )

            split_data = {}
            metadata_dict = {}
            train_indices_list = []

            for key, index in index_result.items():
                split_data[key] = {
                    "train": data.iloc[index["train"]].reset_index(drop=True),
                    "validation": data.iloc[index["validation"]].reset_index(drop=True),
                }

                # Create metadata for both train and validation
                train_metadata = self._create_split_metadata(
                    len(index["train"]), len(index["validation"])
                )
                validation_metadata = self._create_split_metadata(
                    len(index["train"]), len(index["validation"])
                )

                metadata_dict[key] = {
                    "train": train_metadata,
                    "validation": validation_metadata,
                }

                train_indices_list.append(set(index["train"]))

        return split_data, metadata_dict, train_indices_list

    def get_train_indices(self) -> list[set]:
        """
        取得最後一次分割的訓練索引列表，用於向後相容性。

        Returns:
            list[set]: 訓練索引集合列表
        """
        # 這個方法主要用於向後相容，實際使用建議直接使用 split() 的返回值
        if hasattr(self, "_last_train_indices"):
            return self._last_train_indices
        return []

    def _update_metadata_with_split_info(
        self, metadata: SchemaMetadata, train_rows: int, validation_rows: int
    ) -> SchemaMetadata:
        """
        Update metadata with split information using functional approach.

        Args:
            metadata: Original metadata from training data
            train_rows: Number of training rows
            validation_rows: Number of validation rows

        Returns:
            Updated metadata with split information
        """
        # Create new properties with split information
        updated_properties = metadata.properties.copy() if metadata.properties else {}
        updated_properties["row_num_after_split"] = {
            "train": train_rows,
            "validation": validation_rows,
        }

        # Create new metadata instance with updated properties
        return SchemaMetadata(
            schema_id=metadata.schema_id,
            name=metadata.name,
            description=metadata.description,
            fields=metadata.fields,
            properties=updated_properties,
        )

    def _create_split_metadata(
        self, train_rows: int, validation_rows: int
    ) -> SchemaMetadata:
        """
        Create basic metadata for split data.

        Args:
            train_rows: Number of training rows
            validation_rows: Number of validation rows

        Returns:
            Basic metadata with split information
        """
        return SchemaMetadata(
            schema_id="split_data",
            name="Split Data Schema",
            description="Metadata for split data",
            fields=[],  # Will be populated if needed
            properties={
                "row_num_after_split": {
                    "train": train_rows,
                    "validation": validation_rows,
                }
            },
        )

    def _bootstrapping(
        self, index: list, exist_train_indices: list[set] = None
    ) -> dict[int, dict[str, list[int]]]:
        """
        拔靴法生成隨機索引樣本用於資料分割。

        Args:
            index (list): 待分割資料集的索引列表
            exist_train_indices (list[set]): 現有的訓練索引集合列表，用於避免重疊
        """
        if self.config["random_state"] is not None:
            random.seed(self.config["random_state"])

        sample_size = int(len(index) * self.config["train_split_ratio"])

        # 初始化現有訓練索引集合列表
        existing_train_sets = []
        if exist_train_indices:
            existing_train_sets = [
                set(idx_set) if not isinstance(idx_set, set) else idx_set
                for idx_set in exist_train_indices
            ]

        sampled_index = {}

        for n in range(self.config["num_samples"]):
            attempts = 0
            while attempts < self.config["max_attempts"]:
                sampled_indices = set(random.sample(index, sample_size))

                # 檢查是否與現有訓練集合重疊過多
                if self._check_overlap_acceptable(sampled_indices, existing_train_sets):
                    # 將當前樣本加入現有訓練集合列表，供後續比較使用
                    existing_train_sets.append(sampled_indices)

                    sampled_index[n + 1] = {
                        "train": list(sampled_indices),
                        "validation": list(set(index) - sampled_indices),
                    }
                    break

                attempts += 1

            if attempts == self.config["max_attempts"]:
                raise ConfigError(
                    f"Splitter: "
                    f"Unable to sample {self.config['num_samples']} pairs of index "
                    f"with a ratio of {self.config['train_split_ratio']} "
                    f"and max overlap ratio of {self.config['max_overlap_ratio']:.1%} "
                    f"within {self.config['max_attempts']} attempts.\n"
                    f"Consider reducing num_samples, increasing max_overlap_ratio, "
                    f"or increasing max_attempts."
                )
        return sampled_index

    def _check_overlap_acceptable(
        self, new_train_sample: set, existing_train_sets: list[set]
    ) -> bool:
        """
        檢查新訓練樣本與現有訓練集合的重疊是否可接受。

        Args:
            new_train_sample (set): 新的訓練樣本索引集合
            existing_train_sets (list[set]): 現有的訓練索引集合列表

        Returns:
            bool: 如果重疊可接受則返回 True，否則返回 False
        """
        max_overlap_ratio = self.config["max_overlap_ratio"]

        for existing_train_set in existing_train_sets:
            # 1. 檢查是否完全一致
            if new_train_sample == existing_train_set:
                return False

            # 2. 檢查重疊比率是否超過限制
            if max_overlap_ratio < 1.0:  # 只有在不是 100% 時才檢查
                overlap_size = len(new_train_sample.intersection(existing_train_set))
                overlap_ratio = overlap_size / len(new_train_sample)

                if overlap_ratio > max_overlap_ratio:
                    return False

        return True



"""Field layer - Single column management"""

from petsard.metadater.field.field_functions import (
    build_field_metadata,
    calculate_field_stats,
    infer_field_logical_type,
    optimize_field_dtype,
)
from petsard.metadater.field.field_ops import FieldOperations
from petsard.metadater.field.field_types import FieldConfig, FieldMetadata, FieldStats
from petsard.metadater.field.transformation import (
    apply_dtype_conversion,
    transform_field_data,
)
from petsard.metadater.field.type_inference import (
    detect_logical_type_patterns,
    infer_optimal_dtype,
    infer_pandas_dtype,
)

__all__ = [
    "FieldConfig",
    "FieldMetadata",
    "FieldStats",
    "FieldOperations",
    "calculate_field_stats",
    "infer_field_logical_type",
    "optimize_field_dtype",
    "detect_logical_type_patterns",
    "infer_optimal_dtype",
    "infer_pandas_dtype",
    "apply_dtype_conversion",
    "transform_field_data",
]



"""
PETsARD Metadater Module - 元資料管理模組

提供統一的三層架構元資料管理介面：
- Metadata 層: 多表格資料集管理 (datasets)
- Schema 層: 單表格結構管理 (dataframe)
- Field 層: 單欄位分析管理 (column)

主要使用方式：
    from petsard.metadater import Metadater

    # 最常用：分析單表格
    schema = Metadater.create_schema(dataframe, "my_schema")

    # 分析單欄位
    field = Metadater.create_field(series, "my_field")

    # 分析多表格資料集
    metadata = Metadater.analyze_dataset(tables, "my_dataset")
"""

# 主要介面
# SDV adapter 功能
from petsard.metadater.adapters.sdv_adapter import SDVMetadataAdapter
from petsard.metadater.field.field_types import FieldConfig, FieldMetadata

# 核心類型 (使用者需要的)
from petsard.metadater.metadata.metadata_types import Metadata, MetadataConfig
from petsard.metadater.metadater import Metadater
from petsard.metadater.schema.schema_types import SchemaConfig, SchemaMetadata
from petsard.metadater.types.data_types import safe_round

__all__ = [
    # 主要介面
    "Metadater",
    # 核心類型
    "Metadata",
    "MetadataConfig",
    "SchemaMetadata",
    "SchemaConfig",
    "FieldMetadata",
    "FieldConfig",
    # SDV adapter
    "SDVMetadataAdapter",
    # 工具函數
    "safe_round",
]



"""Pure functions for validation operations"""

from typing import Any, Dict, List

import pandas as pd

from petsard.metadater.field.field_types import FieldConfig, FieldMetadata
from petsard.metadater.schema.schema_types import SchemaConfig, SchemaMetadata
from petsard.metadater.types.data_types import DataType, LogicalType


def validate_field_config(config: FieldConfig) -> Dict[str, Any]:
    """
    Pure function to validate field configuration

    Args:
        config: Field configuration to validate

    Returns:
        Validation report dictionary
    """
    violations = []
    warnings = []

    # Validate cast_error values
    if config.cast_error not in ["raise", "coerce", "ignore"]:
        violations.append(
            {
                "type": "invalid_cast_error",
                "value": config.cast_error,
                "message": f"Invalid cast_error value: {config.cast_error}. Must be 'raise', 'coerce', or 'ignore'",
            }
        )

    # Validate logical_type if specified
    if config.logical_type is not None:
        try:
            if isinstance(config.logical_type, str):
                LogicalType(config.logical_type)
            elif not isinstance(config.logical_type, LogicalType):
                violations.append(
                    {
                        "type": "invalid_logical_type",
                        "value": config.logical_type,
                        "message": f"Invalid logical_type: {config.logical_type}",
                    }
                )
        except ValueError:
            violations.append(
                {
                    "type": "invalid_logical_type",
                    "value": config.logical_type,
                    "message": f"Unknown logical_type: {config.logical_type}",
                }
            )

    # Validate type_hint if specified
    if config.type_hint is not None:
        valid_type_hints = [
            "category",
            "datetime",
            "date",
            "time",
            "int",
            "integer",
            "float",
            "string",
            "boolean",
        ]
        if config.type_hint.lower() not in valid_type_hints:
            warnings.append(
                {
                    "type": "unknown_type_hint",
                    "value": config.type_hint,
                    "message": f"Unknown type_hint: {config.type_hint}. Valid hints: {valid_type_hints}",
                }
            )

    return {
        "valid": len(violations) == 0,
        "violations": violations,
        "warnings": warnings,
    }


def validate_schema_config(config: SchemaConfig) -> Dict[str, Any]:
    """
    Pure function to validate schema configuration

    Args:
        config: Schema configuration to validate

    Returns:
        Validation report dictionary
    """
    violations = []
    warnings = []

    # Validate schema_id
    if not config.schema_id or not config.schema_id.strip():
        violations.append(
            {
                "type": "empty_schema_id",
                "message": "schema_id cannot be empty",
            }
        )

    # Validate sample_size
    if config.sample_size is not None and config.sample_size <= 0:
        violations.append(
            {
                "type": "invalid_sample_size",
                "value": config.sample_size,
                "message": f"sample_size must be positive or None, got {config.sample_size}",
            }
        )

    # Validate field configurations
    field_violations = []
    for field_name, field_config in config.fields.items():
        field_validation = validate_field_config(field_config)
        if not field_validation["valid"]:
            field_violations.append(
                {
                    "field": field_name,
                    "violations": field_validation["violations"],
                    "warnings": field_validation["warnings"],
                }
            )

    if field_violations:
        violations.append(
            {
                "type": "invalid_field_configs",
                "field_violations": field_violations,
                "message": f"Invalid field configurations found in {len(field_violations)} fields",
            }
        )

    return {
        "valid": len(violations) == 0,
        "violations": violations,
        "warnings": warnings,
    }


def validate_data_against_field(
    data: pd.Series,
    field_metadata: FieldMetadata,
    strict: bool = False,
) -> Dict[str, Any]:
    """
    Pure function to validate data against field metadata

    Args:
        data: Series to validate
        field_metadata: Field metadata to validate against
        strict: Whether to use strict validation

    Returns:
        Validation report dictionary
    """
    violations = []
    warnings = []

    # Check nullability
    null_count = data.isna().sum()
    if not field_metadata.nullable and null_count > 0:
        violations.append(
            {
                "type": "null_violation",
                "field": field_metadata.name,
                "null_count": int(null_count),
                "total_count": len(data),
                "message": f"Field '{field_metadata.name}' contains {null_count} null values but nullable=False",
            }
        )

    # Check data type compatibility
    if field_metadata.source_dtype:
        expected_dtype = field_metadata.source_dtype
        actual_dtype = str(data.dtype)

        if strict and expected_dtype != actual_dtype:
            violations.append(
                {
                    "type": "dtype_violation",
                    "field": field_metadata.name,
                    "expected": expected_dtype,
                    "actual": actual_dtype,
                    "message": f"Data type mismatch: expected {expected_dtype}, got {actual_dtype}",
                }
            )
        elif expected_dtype != actual_dtype:
            warnings.append(
                {
                    "type": "dtype_mismatch",
                    "field": field_metadata.name,
                    "expected": expected_dtype,
                    "actual": actual_dtype,
                    "message": f"Data type mismatch: expected {expected_dtype}, got {actual_dtype}",
                }
            )

    # Validate logical type constraints
    if field_metadata.logical_type:
        logical_violations = _validate_logical_type_constraints(
            data, field_metadata.logical_type
        )
        violations.extend(logical_violations)

    # Check value ranges for numeric types
    if field_metadata.data_type in [
        DataType.INT8,
        DataType.INT16,
        DataType.INT32,
        DataType.INT64,
        DataType.FLOAT32,
        DataType.FLOAT64,
    ]:
        range_violations = _validate_numeric_ranges(data, field_metadata)
        violations.extend(range_violations)

    return {
        "valid": len(violations) == 0,
        "violations": violations,
        "warnings": warnings,
        "field_name": field_metadata.name,
        "summary": {
            "total_rows": len(data),
            "null_rows": int(null_count),
            "valid_rows": len(data) - int(null_count),
            "violations_count": len(violations),
            "warnings_count": len(warnings),
        },
    }


def validate_dataframe_against_schema(
    data: pd.DataFrame,
    schema: SchemaMetadata,
    strict: bool = False,
) -> Dict[str, Any]:
    """
    Pure function to validate DataFrame against schema

    Args:
        data: DataFrame to validate
        schema: Schema metadata to validate against
        strict: Whether to use strict validation

    Returns:
        Validation report dictionary
    """
    violations = []
    warnings = []
    field_reports = []

    # Check schema-level constraints
    schema_fields = {field.name for field in schema.fields}
    data_fields = set(data.columns)

    missing_fields = schema_fields - data_fields
    extra_fields = data_fields - schema_fields

    if missing_fields:
        violations.append(
            {
                "type": "missing_fields",
                "fields": list(missing_fields),
                "message": f"Missing required fields: {missing_fields}",
            }
        )

    if extra_fields:
        if strict:
            violations.append(
                {
                    "type": "extra_fields",
                    "fields": list(extra_fields),
                    "message": f"Extra fields not allowed in strict mode: {extra_fields}",
                }
            )
        else:
            warnings.append(
                {
                    "type": "extra_fields",
                    "fields": list(extra_fields),
                    "message": f"Extra fields found: {extra_fields}",
                }
            )

    # Validate each field
    for field_metadata in schema.fields:
        if field_metadata.name not in data.columns:
            continue

        field_report = validate_data_against_field(
            data=data[field_metadata.name],
            field_metadata=field_metadata,
            strict=strict,
        )

        field_reports.append(field_report)

        # Aggregate field violations to schema level
        if field_report["violations"]:
            violations.extend(field_report["violations"])
        if field_report["warnings"]:
            warnings.extend(field_report["warnings"])

    return {
        "valid": len(violations) == 0,
        "violations": violations,
        "warnings": warnings,
        "field_reports": field_reports,
        "summary": {
            "total_fields": len(schema.fields),
            "validated_fields": len([r for r in field_reports if r["valid"]]),
            "missing_fields": len(missing_fields),
            "extra_fields": len(extra_fields),
            "total_violations": len(violations),
            "total_warnings": len(warnings),
            "data_shape": data.shape,
        },
    }


def validate_field_metadata_consistency(
    field_metadata: FieldMetadata,
) -> Dict[str, Any]:
    """
    Pure function to validate internal consistency of field metadata

    Args:
        field_metadata: Field metadata to validate

    Returns:
        Validation report dictionary
    """
    violations = []
    warnings = []

    # Check data type and logical type compatibility
    if field_metadata.logical_type:
        compatibility_violations = _check_datatype_logical_type_compatibility(
            field_metadata.data_type, field_metadata.logical_type
        )
        violations.extend(compatibility_violations)

    # Check stats consistency
    if field_metadata.stats:
        stats_violations = _validate_field_stats_consistency(field_metadata.stats)
        violations.extend(stats_violations)

    # Check target dtype compatibility
    if field_metadata.target_dtype and field_metadata.source_dtype:
        if not _are_dtypes_compatible(
            field_metadata.source_dtype, field_metadata.target_dtype
        ):
            warnings.append(
                {
                    "type": "dtype_conversion_warning",
                    "source": field_metadata.source_dtype,
                    "target": field_metadata.target_dtype,
                    "message": f"Conversion from {field_metadata.source_dtype} to {field_metadata.target_dtype} may lose precision",
                }
            )

    return {
        "valid": len(violations) == 0,
        "violations": violations,
        "warnings": warnings,
        "field_name": field_metadata.name,
    }


# Helper functions


def _validate_logical_type_constraints(
    data: pd.Series, logical_type: LogicalType
) -> List[Dict[str, Any]]:
    """Validate data against logical type constraints"""
    violations = []

    if logical_type == LogicalType.EMAIL:
        invalid_emails = _find_invalid_emails(data)
        if invalid_emails:
            violations.append(
                {
                    "type": "invalid_email_format",
                    "invalid_count": len(invalid_emails),
                    "examples": invalid_emails[:5],  # Show first 5 examples
                    "message": f"Found {len(invalid_emails)} invalid email addresses",
                }
            )

    elif logical_type == LogicalType.URL:
        invalid_urls = _find_invalid_urls(data)
        if invalid_urls:
            violations.append(
                {
                    "type": "invalid_url_format",
                    "invalid_count": len(invalid_urls),
                    "examples": invalid_urls[:5],
                    "message": f"Found {len(invalid_urls)} invalid URLs",
                }
            )

    elif logical_type in [LogicalType.LATITUDE, LogicalType.LONGITUDE]:
        invalid_coords = _find_invalid_coordinates(data, logical_type)
        if invalid_coords:
            violations.append(
                {
                    "type": f"invalid_{logical_type.value}",
                    "invalid_count": len(invalid_coords),
                    "examples": invalid_coords[:5],
                    "message": f"Found {len(invalid_coords)} invalid {logical_type.value} values",
                }
            )

    return violations


def _validate_numeric_ranges(
    data: pd.Series, field_metadata: FieldMetadata
) -> List[Dict[str, Any]]:
    """Validate numeric data against expected ranges"""
    violations = []

    if field_metadata.data_type == DataType.INT8:
        out_of_range = data[(data < -128) | (data > 127)].dropna()
    elif field_metadata.data_type == DataType.INT16:
        out_of_range = data[(data < -32768) | (data > 32767)].dropna()
    elif field_metadata.data_type == DataType.INT32:
        out_of_range = data[(data < -2147483648) | (data > 2147483647)].dropna()
    else:
        return violations  # No range check for other types

    if len(out_of_range) > 0:
        violations.append(
            {
                "type": "numeric_range_violation",
                "data_type": field_metadata.data_type.value,
                "out_of_range_count": len(out_of_range),
                "examples": out_of_range.head(5).tolist(),
                "message": f"Found {len(out_of_range)} values outside valid range for {field_metadata.data_type.value}",
            }
        )

    return violations


def _check_datatype_logical_type_compatibility(
    data_type: DataType, logical_type: LogicalType
) -> List[Dict[str, Any]]:
    """Check compatibility between data type and logical type"""
    violations = []

    # Define compatibility rules
    string_logical_types = {
        LogicalType.EMAIL,
        LogicalType.URL,
        LogicalType.UUID,
        LogicalType.PHONE,
        LogicalType.POSTAL_CODE,
        LogicalType.CATEGORICAL,
    }

    numeric_logical_types = {
        LogicalType.LATITUDE,
        LogicalType.LONGITUDE,
        LogicalType.PERCENTAGE,
        LogicalType.CURRENCY,
    }

    if logical_type in string_logical_types and data_type not in [
        DataType.STRING,
        DataType.BINARY,
    ]:
        violations.append(
            {
                "type": "incompatible_data_logical_type",
                "data_type": data_type.value,
                "logical_type": logical_type.value,
                "message": f"Logical type {logical_type.value} requires string data type, got {data_type.value}",
            }
        )

    elif logical_type in numeric_logical_types and data_type not in [
        DataType.INT8,
        DataType.INT16,
        DataType.INT32,
        DataType.INT64,
        DataType.FLOAT32,
        DataType.FLOAT64,
        DataType.DECIMAL,
    ]:
        violations.append(
            {
                "type": "incompatible_data_logical_type",
                "data_type": data_type.value,
                "logical_type": logical_type.value,
                "message": f"Logical type {logical_type.value} requires numeric data type, got {data_type.value}",
            }
        )

    return violations


def _validate_field_stats_consistency(stats) -> List[Dict[str, Any]]:
    """Validate internal consistency of field statistics"""
    violations = []

    # Check basic consistency
    if stats.na_count > stats.row_count:
        violations.append(
            {
                "type": "inconsistent_stats",
                "message": f"na_count ({stats.na_count}) cannot be greater than row_count ({stats.row_count})",
            }
        )

    if stats.distinct_count > stats.row_count:
        violations.append(
            {
                "type": "inconsistent_stats",
                "message": f"distinct_count ({stats.distinct_count}) cannot be greater than row_count ({stats.row_count})",
            }
        )

    # Check percentage calculation
    if stats.row_count > 0:
        expected_na_percentage = (stats.na_count / stats.row_count) * 100
        if (
            abs(stats.na_percentage - expected_na_percentage) > 0.01
        ):  # Allow small rounding errors
            violations.append(
                {
                    "type": "inconsistent_stats",
                    "message": f"na_percentage ({stats.na_percentage}) doesn't match calculated value ({expected_na_percentage:.4f})",
                }
            )

    return violations


def _are_dtypes_compatible(source_dtype: str, target_dtype: str) -> bool:
    """Check if dtype conversion is safe"""
    # Define safe conversions
    safe_conversions = {
        "int8": ["int16", "int32", "int64", "float32", "float64"],
        "int16": ["int32", "int64", "float32", "float64"],
        "int32": ["int64", "float64"],
        "int64": ["float64"],
        "float32": ["float64"],
        "object": ["string", "category"],
    }

    return target_dtype in safe_conversions.get(source_dtype, [target_dtype])


def _find_invalid_emails(data: pd.Series) -> List[str]:
    """Find invalid email addresses in series"""
    import re

    email_pattern = r"^[\w\.-]+@[\w\.-]+\.\w+$"

    invalid_emails = []
    for value in data.dropna():
        if not re.match(email_pattern, str(value)):
            invalid_emails.append(str(value))

    return invalid_emails


def _find_invalid_urls(data: pd.Series) -> List[str]:
    """Find invalid URLs in series"""
    import re

    url_pattern = r"^https?://[^\s]+$"

    invalid_urls = []
    for value in data.dropna():
        if not re.match(url_pattern, str(value)):
            invalid_urls.append(str(value))

    return invalid_urls


def _find_invalid_coordinates(data: pd.Series, coord_type: LogicalType) -> List[float]:
    """Find invalid coordinate values"""
    invalid_coords = []

    for value in data.dropna():
        try:
            coord_value = float(value)
            if coord_type == LogicalType.LATITUDE:
                if not (-90 <= coord_value <= 90):
                    invalid_coords.append(coord_value)
            elif coord_type == LogicalType.LONGITUDE:
                if not (-180 <= coord_value <= 180):
                    invalid_coords.append(coord_value)
        except (ValueError, TypeError):
            invalid_coords.append(value)

    return invalid_coords



import random
from copy import deepcopy

import numpy as np
import pandas as pd

from petsard.exceptions import UnfittedError


class MissingHandler:
    """
    Base class for all MissingHandler classes.
    """

    PROC_TYPE = ("missing",)

    def __init__(self) -> None:
        self._is_fitted: True = False
        self.na_percentage: float = None
        self._imputation_index: list = None
        self._imputation_index_len: int = None
        self.rng = np.random.default_rng()

    def set_na_percentage(self, na_percentage: float = 0.0) -> None:
        """
        Set NA percentage for the instance.

        Args:
            na_percentage (float, default=0.0): NA percentage from the metadata.
        """
        if na_percentage > 1.0 or na_percentage < 0.0:
            raise ValueError("Invalid NA percentage. It should be between 0.0 and 1.0.")

        self.na_percentage = na_percentage

    def set_imputation_index(self, index_list: list = []) -> None:
        """
        Determine which indices can be imputed as NA globally.

        Args:
            index_list (list): The indices can be set to NA.
        """
        if not isinstance(index_list, list):
            raise ValueError("Invalid index_list. It should be a list.")

        self._imputation_index = index_list
        self._imputation_index_len = len(index_list)

    def fit(self, data: pd.Series) -> None:
        """
        Base method of `fit`.

        Args:
            data (pd.Series): The data needed to be fitted.
        """
        self._fit(data)

        self._is_fitted = True

    def _fit():
        """
        _fit method is implemented in subclasses.

        fit method is responsible for general action defined by the base class.
        _fit method is for specific procedure conducted by each subclasses.
        """
        raise NotImplementedError(
            "_fit method should be implemented " + "in subclasses."
        )

    def transform(self, data: pd.Series) -> pd.Series | np.ndarray:
        """
        Base method of `transform`.

        Args:
            data (pd.Series): The data needed to be transformed.

        Return:
            (pd.Series | np.ndarray): The transformed data.
        """
        # Check the object is fitted
        if not self._is_fitted:
            raise UnfittedError("The object is not fitted. Use .fit() first.")

        return self._transform(data)

    def _transform():
        """
        _transform method is implemented in subclasses.

        transform method is responsible for general action
            defined by the base class.
        _transform method is for specific procedure
            conducted by each subclasses.
        """
        raise NotImplementedError(
            "_transform method should be implemented " + "in subclasses."
        )

    def inverse_transform(self, data: pd.Series) -> pd.Series:
        """
        Insert NA into the data to have the same pattern with the original data.

        Args:
            data (pd.Series): The data needed to be transformed inversely.

        Return:
            (pd.Series): The inverse transformed data.
        """
        # Check the object is fitted
        if not self._is_fitted:
            raise UnfittedError("The object is not fitted. Use .fit() first.")

        if self.na_percentage == 0.0 or self._imputation_index_len == 0:
            return data
        else:
            _na_mask = self.rng.choice(
                self._imputation_index,
                size=int(self.na_percentage * self._imputation_index_len),
                replace=False,
            )
            _col_data = deepcopy(data)
            _col_data.iloc[_na_mask] = np.nan

            return _col_data

    def _inverse_transform():
        """
        _inverse_transform method is implemented in subclasses.

        inverse_transform method is responsible for general action
            defined by the base class.
        _inverse_transform method is for specific procedure
            conducted by each subclasses.
        """
        raise NotImplementedError(
            "_inverse_transform method should be " + "implemented in subclasses."
        )


class MissingMean(MissingHandler):
    """
    Impute NA values with the mean value.
    """

    def __init__(self) -> None:
        super().__init__()
        self.data_mean: float = None

    def _fit(self, data: pd.Series) -> None:
        """
        Gather information for transformation and reverse transformation.

        Args:
            data (pd.Series): The data needed to be transformed.
        """

        self.data_mean = data.mean()

    def _transform(self, data: pd.Series) -> pd.Series:
        """
        Fill NA with mean.

        Args:
            data (pd.Series): The data needed to be transformed.

        Return:
            (pd.Series): The transformed data.
        """

        return data.fillna(self.data_mean)

    def _inverse_transform(self, data: None) -> None:
        pass  # Redundant


class MissingMedian(MissingHandler):
    """
    Impute NA values with the median value.
    """

    def __init__(self) -> None:
        super().__init__()
        self.data_median: float = None

    def _fit(self, data: pd.Series) -> None:
        """
        Gather information for transformation and reverse transformation.

        Args:
            data (pd.Series): The data needed to be transformed.
        """

        self.data_median = data.median()

    def _transform(self, data: pd.Series) -> pd.Series:
        """
        Fill NA with median.

        Args:
            data (pd.Series): The data needed to be transformed.

        Return:
            (pd.Series): The transformed data.
        """

        return data.fillna(self.data_median)

    def _inverse_transform(self, data: None) -> None:
        pass  # Redundant


class MissingSimple(MissingHandler):
    """
    Impute NA values with the given value.
    """

    def __init__(self, value: float = 0.0) -> None:
        """
        Args:
            value (float, default=0.0): The value for imputation.
        """
        super().__init__()
        self.data_value: float = value

    def _fit(self, data: None) -> None:
        pass  # Redundant

    def _transform(self, data: pd.Series) -> pd.Series:
        """
        Fill NA with median.

        Args:
            data (pd.Series): The data needed to be transformed.

        Return:
            (pd.Series): The transformed data.
        """

        return data.fillna(self.data_value)

    def _inverse_transform(self, data: None) -> None:
        pass  # Redundant


class MissingDrop(MissingHandler):
    """
    Drop the rows with NA values.
    """

    def __init__(self) -> None:
        super().__init__()
        self.data_backup: pd.Series = None  # for restoring data

    def _fit(self, data: None) -> None:
        pass  # Redundant

    def _transform(self, data: pd.Series) -> np.ndarray:
        """
        Mark the NA cells and store the original data.

        Args:
            data (pd.Series): The data needed to be transformed.

        Return:
            (np.ndarray): The filter marking the NA cells.
        """
        self.data_backup = data

        return data.isna().values.ravel()

    def _inverse_transform(self, data: None) -> None:
        pass  # Redundant


class MissingMode(MissingHandler):
    """
    Impute NA values with the mode value.
    """

    def __init__(self) -> None:
        super().__init__()
        self.data_mode: list[str] | list[int] | list[float] = None

    def _fit(self, data: pd.Series) -> None:
        """
        Gather information for transformation and reverse transformation.

        Args:
            data (pd.Series): The data needed to be transformed.
        """
        value_counts: pd.Series = data.value_counts()
        max_count = value_counts.max()
        self.data_mode = value_counts[value_counts == max_count].index.tolist()

    def _transform(self, data: pd.Series) -> pd.Series:
        """
        Fill NA with mode.

        Args:
            data (pd.Series): The data needed to be transformed.

        Return:
            (pd.Series): The transformed data.
        """
        if len(self.data_mode) == 1:
            return data.fillna(self.data_mode[0])
        else:
            filled = data.copy()
            for idx, _ in filled[filled.isna()].items():
                filled[idx] = random.choice(self.data_mode)
            return filled

    def _inverse_transform(self, data: None) -> None:
        pass  # Redundant



import logging
import re
import warnings
from dataclasses import dataclass
from enum import Enum, auto
from typing import Any, Union

import pandas as pd
from sdmetrics.reports.base_report import BaseReport
from sdmetrics.reports.single_table import DiagnosticReport, QualityReport
from sdv.metadata import Metadata as SDV_Metadata

from petsard.config_base import BaseConfig
from petsard.evaluator.evaluator_base import BaseEvaluator
from petsard.exceptions import UnsupportedMethodError
from petsard.metadater import safe_round


class SDMetricsSingleTableMap(Enum):
    """
    Mapping of SDMetrics.
    """

    DIAGNOSTICREPORT: int = auto()
    QUALITYREPORT: int = auto()

    @classmethod
    def map(cls, method: str) -> int:
        """
        Get suffixes mapping int value
            Accept both of "sdmetrics-" or "sdmetrics-single_table-" prefix

        Args:
            method (str): evaluating method

        Return:
            (int): The method code.
        """
        return cls.__dict__[
            re.sub(r"^(sdmetrics-single_table-|sdmetrics-)", "", method).upper()
        ]


@dataclass
class SDMetricsSingleTableConfig(BaseConfig):
    """
    Configuration for the sdmetrics single-table evaluator.

    Attributes:
        _logger (logging.Logger): The logger object.
        real_data (pd.DataFrame): The real data. copy from
        synthetic_data (pd.DataFrame): The synthetic data.
        metadata (dict): The metadata of the data.
        ori (pd.DataFrame): The original data.
        syn (pd.DataFrame): The synthetic data.
    """

    ori: pd.DataFrame
    syn: pd.DataFrame
    real_data: pd.DataFrame = None
    synthetic_data: pd.DataFrame = None
    metadata: dict = None

    def __post_init__(self):
        super().__post_init__()

        if not all(isinstance(attr, pd.DataFrame) for attr in [self.ori, self.syn]):
            error_msg: str = "The 'ori' and 'syn' attributes must be DataFrames."
            self._logger.error(error_msg)
            raise TypeError(error_msg)

        self.real_data = self.ori.copy()
        self.synthetic_data = self.syn.copy()
        self.ori = None
        self.syn = None

    def create_metadata(self) -> None:
        """
        Create metadata from the original data.
        """
        self._logger.debug("Creating SDV metadata")
        sdv_metadata_result: SDV_Metadata = SDV_Metadata().detect_from_dataframe(
            self.real_data
        )
        self.metadata = sdv_metadata_result._convert_to_single_table().to_dict()


class SDMetricsSingleTable(BaseEvaluator):
    """
    Factory for SDMetrics Evaluator.
    """

    REQUIRED_INPUT_KEYS: list[str] = ["ori", "syn"]
    SDMETRICS_SINGLETABLE_CLASS_MAP: dict[int, BaseReport] = {
        SDMetricsSingleTableMap.DIAGNOSTICREPORT: DiagnosticReport,
        SDMetricsSingleTableMap.QUALITYREPORT: QualityReport,
    }
    REQUIRED_SDMETRICS_SINGLETABLE_KEYS: list[str] = [
        "real_data",
        "synthetic_data",
        "metadata",
    ]
    AVAILABLE_SCORES_GRANULARITY: list[str] = [
        "global",
        "columnwise",
        "pairwise",
    ]
    SDMETRICS_SCORES_GRANULARITY_PROPERTIES_MAP: dict[tuple[str, int], str] = {
        ("columnwise", SDMetricsSingleTableMap.DIAGNOSTICREPORT): "Data Validity",
        ("columnwise", SDMetricsSingleTableMap.QUALITYREPORT): "Column Shapes",
        ("pairwise", SDMetricsSingleTableMap.QUALITYREPORT): "Column Pair Trends",
    }

    def __init__(self, config: dict):
        """
        Args:
            config (dict): A dictionary containing the configuration settings.
                - eval_method (str): The method of how you evaluating data.

        Attributes:
            REQUIRED_INPUT_KEYS (list[str]): The required input keys.
            SDMETRICS_SINGLETABLE_CLASS_MAP (dict[int, BaseReport]): The mapping of the SDMetrics classes.
            REQUIRED_SDMETRICS_SINGLETABLE_KEYS (list[str]): The required SDMetrics keys.
            AVAILABLE_SCORES_GRANULARITY (list[str]): The available scores granularity.
            SDMETRICS_SCORES_GRANULARITY_PROPERTIES_MAP (dict[tuple[str, int], str]): The mapping of the scores granularity properties.
            _logger (logging.Logger): The logger object.
            config (dict): A dictionary containing the configuration settings.
            _impl (Any): The evaluator object.
        """
        super().__init__(config=config)
        self._logger: logging.Logger = logging.getLogger(
            f"PETsARD.{self.__class__.__name__}"
        )

        self._logger.debug(
            f"Initializing synthesizer with method: {self.config['eval_method']}"
        )
        try:
            method_code: int = SDMetricsSingleTableMap.map(self.config["eval_method"])
            self._logger.debug(f"Mapped method code: {method_code}")
            evaluator_class: Any = self.SDMETRICS_SINGLETABLE_CLASS_MAP[method_code]
        except KeyError:
            error_msg: str = (
                f"Unsupported evaluator method: {self.config['eval_method']}"
            )
            self._logger.error(error_msg)
            raise UnsupportedMethodError(error_msg)

        self._impl: BaseReport = evaluator_class()

    def _extract_scores(self) -> dict[str, pd.DataFrame]:
        """
        _extract_result of SDMetrics.
            Uses .get_score()/.get_properties()/.get_details() method in SDMetrics
            to extract result from self._impl into the designated dictionary.

        Return
            (dict[str, pd.DataFrame]) Result as following key-value pairs:
                - score
                - properties
                - details
        """

        sdmetrics_scores: dict[str, pd.DataFrame] = {}

        self._logger.debug("Extract scores level from SDMetrics")
        sdmetrics_scores["score"] = safe_round(self._impl.get_score())

        # Tranfer pandas to desired dict format:
        #     {'properties name': {'Score': ...}, ...}
        properties = self._impl.get_properties()
        properties["Score"] = safe_round(properties["Score"])

        self._logger.debug("Extracting properties level from SDMetrics")
        sdmetrics_scores["properties"] = (
            properties.set_index("Property").rename_axis(None).to_dict("index")
        )

        self._logger.debug("Extracting details level from SDMetrics")
        sdmetrics_scores["details"] = {}
        for property in sdmetrics_scores["properties"].keys():
            sdmetrics_scores["details"][property] = self._impl.get_details(
                property_name=property
            )

        return sdmetrics_scores

    def _get_global(
        self, sdmetrics_scores: dict[str, pd.DataFrame]
    ) -> Union[pd.DataFrame, None]:
        """
        Returns the global result from the SDMetrics.

        Args:
            sdmetrics_scores (dict[str, pd.DataFrame]): The SDMetrics scores.
            property (str): The name of the property.

        Returns:
            (pd.DataFrame): A DataFrame with the global evaluation result.
                One row only for representing the whole data result.
        """
        # get_score
        data = {"Score": sdmetrics_scores["score"]}
        # get_properties
        data.update(
            {
                key: value["Score"]
                for key, value in sdmetrics_scores["properties"].items()
            }
        )
        return pd.DataFrame.from_dict(data={"result": data}, orient="columns").T

    def _transform_details(
        self,
        sdmetrics_scores: dict[str, pd.DataFrame],
        property: str,
    ) -> pd.DataFrame:
        """
        Transforms the details of a specific property in the result dictionary.

        Args:
            sdmetrics_scores (dict[str, pd.DataFrame]): The SDMetrics scores.
            property (str): The name of the property.

        Returns:
            (pd.DataFrame) The transformed details dataframe.
        """
        data: pd.DataFrame = sdmetrics_scores["details"][property].copy()

        # set column as index, and remove index name
        if "Column" in data.columns:
            data.set_index("Column", inplace=True)
            data.index.name = None
        else:
            # set pairwise columns as one column
            data.set_index(["Column 1", "Column 2"], inplace=True)
            data.index.names = [None, None]

        # set Property
        data["Property"] = property

        # sort columns
        return data[
            ["Property", "Metric"]
            + [col for col in data.columns if col not in ["Property", "Metric"]]
        ]

    def _eval(self, data: dict[str, pd.DataFrame]) -> dict[str, pd.DataFrame]:
        """
        Evaluating the evaluator.
            _impl should be initialized in this method.

        Args:
            data (dict[str, pd.DataFrame]): The data to be evaluated.

        Return:
            (dict[str, pd.DataFrame]): The evaluation result
        """
        self._logger.info("Evaluating with data")

        self._logger.debug("Initializing SDMetricsSingleTableConfig")
        sdmetrics_singletable_config: SDMetricsSingleTableConfig = (
            SDMetricsSingleTableConfig.from_dict(data)
        )
        sdmetrics_singletable_config.create_metadata()

        self._logger.debug("Initializing evaluator in _eval method")

        # catch warnings during synthesizer initialization:
        # "We strongly recommend saving the metadata using 'save_to_json' for replicability in future SDV versions."
        with warnings.catch_warnings(record=True) as w:
            warnings.simplefilter("always")
            self._impl.generate(
                **sdmetrics_singletable_config.get_params(
                    param_configs=[
                        {attr: {"action": "INCLUDE"}}
                        for attr in self.REQUIRED_SDMETRICS_SINGLETABLE_KEYS
                    ]
                )
            )

            for warning in w:
                self._logger.debug(f"Warning during _eval: {warning.message}")

        self._logger.info("Successfully evaluating from data")

        sdmetrics_scores: dict[str, pd.DataFrame] = self._extract_scores()
        self._logger.debug(f"Extracted scores: {list(sdmetrics_scores.keys())}")

        scores: dict[str, pd.DataFrame] = {}
        property: str = None
        for granularity in self.AVAILABLE_SCORES_GRANULARITY:
            self._logger.debug(f"Extracting {granularity} level as PETsARD format")

            if granularity == "global":
                scores[granularity] = self._get_global(
                    sdmetrics_scores=sdmetrics_scores,
                )
            else:
                method_code: int = SDMetricsSingleTableMap.map(
                    self.config["eval_method"]
                )
                if (
                    granularity,
                    method_code,
                ) in self.SDMETRICS_SCORES_GRANULARITY_PROPERTIES_MAP:
                    property = self.SDMETRICS_SCORES_GRANULARITY_PROPERTIES_MAP[
                        (
                            granularity,
                            method_code,
                        )
                    ]

                if property is None:
                    self._logger.debug(
                        f"Property not found for {granularity} level. Skipping."
                    )
                    scores[granularity] = None
                else:
                    scores[granularity] = self._transform_details(
                        sdmetrics_scores, property=property
                    )
        self._logger.info("Successfully extracting scores")

        return scores



import logging
import os
import subprocess
import sys
from pathlib import Path
from typing import Any, Optional, Tuple, Type

import requests

from petsard.utils import load_external_module


def setup_environment(
    is_colab: bool,
    branch: str = "main",
    benchmark_data: list[str] = None,
    example_files: list[str] = None,
    subfolder: Optional[str] = None,
) -> None:
    """
    Setup the environment for both Colab and local development

    Args:
        is_colab (bool): Whether running in Colab environment
        branch (str, optional): The GitHub branch to use, defaults to "main"
        benchmark_data (list[str], optional):
            The dataset list of benchmark data to load by PETsARD Loader
        example_files (list[str], optional):
            List of example files to download from GitHub
        subfolder (str, optional):
            The subfolder in the demo directory. Defaults to None
    """
    # Check Python version
    if sys.version_info < (3, 10):
        raise RuntimeError(
            "Requires Python 3.10+, "
            f"current version is {sys.version_info.major}.{sys.version_info.minor}"
        )

    # Ensure pip is installed
    subprocess.run([sys.executable, "-m", "ensurepip"], check=True)
    # avoid pip version warning
    os.environ["PIP_DISABLE_PIP_VERSION_CHECK"] = "1"

    if is_colab:
        # Install petsard directly from GitHub
        subprocess.run(
            [
                sys.executable,
                "-m",
                "pip",
                "install",
                f"git+https://github.com/nics-tw/petsard.git@{branch}#egg=petsard",
            ],
            check=True,
        )
        from IPython.display import clear_output

        clear_output(wait=True)
    else:
        # Find the project root directory
        demo_dir = Path.cwd()

        # Calculate project root based on current directory structure
        if subfolder:
            # If we're in a subfolder like demo/use-cases/preproc/
            # we need to go up to find the project root (where pyproject.toml is)
            current_path = demo_dir
            project_root = None

            # Search upwards for pyproject.toml
            for parent in [current_path] + list(current_path.parents):
                if (parent / "pyproject.toml").exists():
                    project_root = parent
                    break

            if project_root is None:
                raise FileNotFoundError(
                    "Could not find project root with pyproject.toml"
                )
        else:
            # If we're directly in demo/, project root is parent
            project_root = demo_dir.parent

        # Local installation
        subprocess.run(
            [sys.executable, "-m", "pip", "install", "-e", str(project_root)],
            check=True,
        )

    print("Installation complete!")

    if benchmark_data:
        from petsard.loader import Loader

        for benchmark in benchmark_data:
            try:
                loader = Loader(filepath=f"benchmark://{benchmark}")
                loader.load()
                print(f"Successful loading benchmark data: {benchmark}")
            except Exception as e:
                print(f"Failed to loading {benchmark}: {e}")

    # Download example files if specified
    if example_files and is_colab:
        for repo_path in example_files:
            # Get just the filename for local path
            local_file = Path(repo_path).name

            # Construct GitHub raw content URL
            file_url = f"https://raw.githubusercontent.com/nics-tw/petsard/{branch}/{repo_path}"

            try:
                response = requests.get(file_url)
                response.raise_for_status()

                # Write to current directory
                with open(local_file, "w") as f:
                    f.write(response.text)
                print(f"Successfully downloaded: {repo_path} -> {local_file}")
            except Exception as e:
                print(f"Failed to download {repo_path}: {e}")


def get_yaml_path(
    is_colab: bool,
    yaml_file: str,
    branch: str = "main",
    subfolder: Optional[str] = None,
) -> Path:
    """
    Get the YAML file path and display its content,
        supporting both Colab and local environments

    Args:
        is_colab (bool): Whether running in Colab environment
        yaml_file (str): Name of the YAML file
        branch (str, optional):
            The branch name to fetch YAML from GitHub. Defaults to "main"
        subfolder (str, optional):
            The subfolder in the demo directory. Defaults to None

    Returns:
        Path: Path to the YAML file

    Raises:
        FileNotFoundError: When file not found in local environment
        requests.RequestException: When failed to download file in Colab
    """
    if is_colab:
        import tempfile

        import requests

        yaml_url = (
            "https://raw.githubusercontent.com/nics-tw/"
            f"petsard/{branch}/yaml/{yaml_file}"
        )

        response = requests.get(yaml_url)
        if response.status_code != 200:
            raise requests.RequestException(
                f"Failed to download YAML file. Status code: {response.status_code}, URL: {yaml_url}"
            )

        with tempfile.NamedTemporaryFile(
            mode="w", suffix=".yaml", delete=False
        ) as tmp_file:
            tmp_file.write(response.text)
            print("Configuration content:\n\n")
            print(response.text)
            return Path(tmp_file.name)
    else:
        demo_dir = Path.cwd()

        # Calculate project root based on current directory structure
        if subfolder:
            # Search upwards for pyproject.toml to find project root
            current_path = demo_dir
            project_root = None

            for parent in [current_path] + list(current_path.parents):
                if (parent / "pyproject.toml").exists():
                    project_root = parent
                    break

            if project_root is None:
                raise FileNotFoundError(
                    "Could not find project root with pyproject.toml"
                )

            yaml_path = project_root / "demo" / subfolder / yaml_file
        else:
            # If we're directly in demo/, project root is parent
            project_root = demo_dir.parent
            yaml_path = project_root / "demo" / yaml_file

        if not yaml_path.exists():
            raise FileNotFoundError(
                f"YAML file not found at {yaml_path}. "
                "Please make sure you have forked the project "
                "and are in the correct directory"
            )

        with open(yaml_path) as f:
            content = f.read()
            print("Configuration content:")
            print(content)

        return yaml_path


def load_demo_module(
    module_path: str,
    class_name: str,
    logger: logging.Logger,
    required_methods: dict[str, list[str]] = None,
) -> Tuple[Any, Type]:
    """
    Load external Python module for demo purposes with intelligent path resolution.

    This function is specifically designed for demo environments and supports
    multi-level subdirectory searching within the demo folder structure.

    Args:
        module_path (str): Path to the external module (relative or absolute)
        class_name (str): Name of the class to load from the module
        logger (logging.Logger): Logger for recording messages
        required_methods (Dict[str, List[str]], optional):
            Dictionary mapping method names to required parameter names
            e.g. {"fit": ["data"], "sample": []}

    Returns:
        Tuple[Any, Type]: A tuple containing the module instance and the class

    Raises:
        FileNotFoundError: If the module file does not exist
        ConfigError: If the module cannot be loaded or doesn't contain the specified class
    """
    # Create demo-specific search paths
    demo_search_paths = _get_demo_search_paths(module_path)

    # Use the core petsard function with demo search paths
    return load_external_module(
        module_path, class_name, logger, required_methods, demo_search_paths
    )


def _get_demo_search_paths(module_path: str) -> list[str]:
    """
    Get demo-specific search paths for module resolution.

    Args:
        module_path (str): The module path to create search paths for

    Returns:
        List[str]: List of demo-specific search paths
    """
    # Get the current working directory
    cwd = os.getcwd()

    # Demo-specific search locations
    demo_search_paths = [
        # 3. Relative to current working directory with demo prefix
        os.path.join(cwd, "demo", module_path),
        # 4. Search in demo subdirectories for multi-level support
        os.path.join(cwd, "demo", "use-cases", module_path),
        os.path.join(cwd, "demo", "use-cases", "preproc", module_path),
        os.path.join(cwd, "demo", "best-practices", module_path),
        os.path.join(cwd, "demo", "developer-guide", module_path),
    ]

    return demo_search_paths



import pandas as pd

from petsard.evaluator.evaluator_base import BaseEvaluator
from petsard.exceptions import ConfigError
from petsard.utils import load_external_module


class CustomEvaluator(BaseEvaluator):
    """
    Adapter class for Custom evaluator as method = 'custom_method'
    """

    REQUIRED_METHODS: dict[str, str] = {
        "__init__": ["config"],
        "eval": ["data"],
    }

    def __init__(self, config: dict):
        """
        Args:
            config (dict): The configuration assign by Synthesizer

        Attributes:
            _logger (logging.Logger): The logger object.
            config (dict): The configuration of the evaluator_base.
            _impl (Any): The evaluator
            REQUIRED_INPUT_KEYS (list[str]): The required input keys for the evaluator.
            AVAILABLE_SCORES_GRANULARITY (list[str]): The available scores granularity.
        """
        super().__init__(config)

        if "module_path" not in self.config:
            error_msg: str = (
                "Module path (module_path) is not provided in the configuration."
            )
            self._logger.error(error_msg)
            raise ConfigError(error_msg)

        if "class_name" not in self.config:
            error_msg: str = (
                "Class name (class_name) is not provided in the configuration."
            )
            self._logger.error(error_msg)
            raise ConfigError(error_msg)

        evaluator_class: callable = None

        # Use demo-specific module loading for better path resolution
        try:
            from demo.utils import load_demo_module

            _, evaluator_class = load_demo_module(
                module_path=self.config["module_path"],
                class_name=self.config["class_name"],
                logger=self._logger,
                required_methods=self.REQUIRED_METHODS,
            )
        except ImportError:
            # Fallback to core function if demo utils not available
            _, evaluator_class = load_external_module(
                module_path=self.config["module_path"],
                class_name=self.config["class_name"],
                logger=self._logger,
                required_methods=self.REQUIRED_METHODS,
            )

        self.REQUIRED_INPUT_KEYS: list[str] = evaluator_class.REQUIRED_INPUT_KEYS
        self.AVAILABLE_SCORES_GRANULARITY: list[str] = (
            evaluator_class.AVAILABLE_SCORES_GRANULARITY
        )

        self._impl = evaluator_class(config=config)

    def _eval(self, data: dict[str, pd.DataFrame]) -> dict[str, pd.DataFrame]:
        """
        Evaluating the evaluator.
            _impl should be initialized in this method.

        Args:
            data (dict[str, pd.DataFrame]): The data to be evaluated.

        Return:
            (dict[str, pd.DataFrame]): The evaluation result
        """
        return self._impl.eval(data=data)



import logging
from dataclasses import dataclass, field
from enum import Enum, auto
from typing import Any, Optional

import numpy as np
import pandas as pd

from petsard.config_base import BaseConfig
from petsard.evaluator.evaluator_base import BaseEvaluator
from petsard.evaluator.stats_base import (
    StatsJSDivergence,
    StatsMax,
    StatsMean,
    StatsMedian,
    StatsMin,
    StatsNUnique,
    StatsStd,
)
from petsard.exceptions import ConfigError, UnsupportedMethodError
from petsard.metadater import Metadater, safe_round


class StatsMap(Enum):
    """
    Mapping of the statistics method to the corresponding code.
    """

    MEAN = auto()
    STD = auto()
    MEDIAN = auto()
    MIN = auto()
    MAX = auto()
    NUNIQUE = auto()
    JSDIVERGENCE = auto()

    @classmethod
    def map(cls, method: str) -> int:
        """
        Get suffixes mapping int value
            Accept both of "sdmetrics-" or "sdmetrics-single_table-" prefix

        Args:
            method (str): evaluating method

        Return:
            (int): The method code.
        """
        return cls.__dict__[method.upper()]


@dataclass
class StatsConfig(BaseConfig):
    """
    Configuration for the Stats Evaluator.

    Attributes:
        eval_method (str): The evaluation method.
        eval_method_code (int): The evaluation method code.
        AVALIABLE_STATS_METHODS (list[str]): The available statistics methods.
        AVAILABLE_COMPARE_METHODS (list[str]): The available compare methods.
        AVAILABLE_AGGREGATED_METHODS (list[str]): The available aggregated methods.
        AVAILABLE_SUMMARY_METHODS (list[str]): The available summary methods.
        REQUIRED_INPUT_KEYS (list[str]): The required input keys.
        stats_method (list[str]): The statistics methods.
        compare_method (str): The compare method.
        aggregated_method (str): The aggregated method.
        summary_method (str): The summary method.
        columns_info (dict): The columns information.
            - ori/syn_dtype (str): The data type of the column.
            - ori/syn_infer_dtype (str): The inferred data type of the column.
            - dtype_match (bool): Whether the data type matches.
            - infer_dtype_match (bool): Whether the inferred data type matches.

    """

    eval_method: str
    eval_method_code: Optional[int] = None
    AVAILABLE_STATS_METHODS: list[str] = field(
        default_factory=lambda: [
            "mean",
            "std",
            "median",
            "min",
            "max",
            "nunique",
            "jsdivergence",
        ]
    )
    AVAILABLE_COMPARE_METHODS: list[str] = field(
        default_factory=lambda: ["diff", "pct_change"]
    )
    AVAILABLE_AGGREGATED_METHODS: list[str] = field(default_factory=lambda: ["mean"])
    AVAILABLE_SUMMARY_METHODS: list[str] = field(default_factory=lambda: ["mean"])
    REQUIRED_INPUT_KEYS: list[str] = field(default_factory=lambda: ["ori", "syn"])
    stats_method: list[str] = field(
        default_factory=lambda: [
            "mean",
            "std",
            "median",
            "min",
            "max",
            "nunique",
            "jsdivergence",
        ]
    )
    compare_method: str = "pct_change"
    aggregated_method: str = "mean"
    summary_method: str = "mean"
    columns_info: Optional[dict[str, dict[str, str]]] = field(default_factory=dict)

    def __post_init__(self):
        super().__post_init__()
        error_msg: Optional[str] = None

        invalid_methods: list[str] = [
            method
            for method in self.stats_method
            if method not in self.AVAILABLE_STATS_METHODS
        ]
        if invalid_methods:
            error_msg = (
                f"Invalid stats method: {invalid_methods}. "
                f"Available methods are: {self.AVAILABLE_STATS_METHODS}"
            )
            self._logger.error(error_msg)
            raise UnsupportedMethodError(error_msg)

        if self.compare_method not in self.AVAILABLE_COMPARE_METHODS:
            error_msg = f"Invalid compare method: {self.compare_method}."
            self._looger.error(error_msg)
            raise UnsupportedMethodError(error_msg)

        if self.aggregated_method not in self.AVAILABLE_AGGREGATED_METHODS:
            error_msg = f"Invalid aggregated method: {self.aggregated_method}."
            self._logger.error(error_msg)
            raise UnsupportedMethodError(error_msg)

        if self.summary_method not in self.AVAILABLE_SUMMARY_METHODS:
            error_msg = f"Invalid summary method: {self.summary_method}."
            self._logger.error(error_msg)
            raise UnsupportedMethodError(error_msg)

    def update_data(self, data: dict[str, pd.DataFrame]) -> None:
        error_msg: Optional[str] = None

        self._logger.info(
            f"Updating data with {len(self.REQUIRED_INPUT_KEYS)} required keys"
        )

        # Validate required keys
        if not all(key in data for key in self.REQUIRED_INPUT_KEYS):
            error_msg = f"Missing required keys. Expected: {self.REQUIRED_INPUT_KEYS}"
            self._logger.error(error_msg)
            raise ConfigError(error_msg)

        ori_colnames: list[str] = data["ori"].columns
        syn_colnames: list[str] = data["syn"].columns
        colnames = list(set(ori_colnames) & set(syn_colnames))
        self._logger.debug(
            f"Found {len(colnames)} common columns between original and synthetic data"
        )

        self.columns_info = {}
        temp: dict[str, str] = None
        dtype = None
        for colname in colnames:
            temp = {}

            for source in ["ori", "syn"]:
                if colname in data[source].columns:
                    dtype = data[source][colname].dtype
                    # Create a temporary series to analyze the data type
                    temp_series = data[source][colname]
                    field_metadata = Metadater.create_field(
                        series=temp_series,
                        field_name=colname,
                        compute_stats=False,
                        infer_logical_type=False,
                        optimize_dtype=False,
                    )
                    temp.update(
                        {
                            f"{source}_dtype": dtype,
                            f"{source}_infer_dtype": field_metadata.data_type.value.lower(),
                        }
                    )
            temp["dtype_match"] = temp.get("ori_dtype") == temp.get("syn_dtype")
            temp["infer_dtype_match"] = temp.get("ori_infer_dtype") == temp.get(
                "syn_infer_dtype"
            )
            self.columns_info[colname] = temp
        self._logger.debug(
            f"Column information updated for {len(self.columns_info)} columns"
        )


class Stats(BaseEvaluator):
    """
    Evaluator for Statistics method.
    """

    REQUIRED_INPUT_KEYS: list[str] = ["ori", "syn"]
    INFER_DTYPE_MAP: dict[str, str] = {
        method: dtype
        for dtype, methods in {
            "numerical": [
                "mean",
                "std",
                "median",
                "min",
                "max",
            ],
            "categorical": ["nunique", "jsdivergence"],
        }.items()
        for method in methods
    }
    EXEC_GRANULARITY_MAP: dict[str, str] = {
        method: granularity
        for granularity, methods in {
            "columnwise": [
                "mean",
                "std",
                "median",
                "min",
                "max",
                "nunique",
            ],
            "percolumn": [
                "jsdivergence",
            ],
        }.items()
        for method in methods
    }
    MODULE_MAP: dict[str, callable] = {
        "mean": StatsMean,
        "std": StatsStd,
        "median": StatsMedian,
        "min": StatsMin,
        "max": StatsMax,
        "nunique": StatsNUnique,
        "jsdivergence": StatsJSDivergence,
    }
    COMPARE_METHOD_MAP: dict[str, dict[str, callable]] = {
        "diff": {
            "func": lambda syn, ori: syn - ori,
            "handle_zero": lambda syn, ori: syn - ori,
        },
        "pct_change": {
            "func": lambda syn, ori: (syn - ori) / abs(ori),
            "handle_zero": lambda syn, ori: np.nan,
        },
    }
    AGGREGATED_METHOD_MAP: dict[str, callable] = {
        "mean": lambda df: {k: safe_round(v) for k, v in df.mean().to_dict().items()}
    }
    SUMMARY_METHOD_MAP: dict[str, callable] = {
        "mean": lambda values: safe_round(np.mean(list(values)))
    }
    AVAILABLE_SCORES_GRANULARITY: list[str] = ["global", "columnwise"]  # "pairwise"

    def __init__(self, config: dict):
        """
        Args:
            config (dict): The configuration assign by Evaluator.
        """
        super().__init__(config=config)
        self._logger: logging.Logger = logging.getLogger(
            f"PETsARD.{self.__class__.__name__}"
        )

        self._logger.debug(f"Verifying StatsConfig with parameters: {self.config}")
        self.stats_config = StatsConfig(**self.config)
        self._logger.debug("StatsConfig successfully initialized")

        self._impl: Optional[dict[str, callable]] = None

    def _process_columnwise(
        self, data: dict[str, pd.DataFrame], col: str, info: dict[str, Any], method: str
    ) -> dict[str, float]:
        """
        Process column-wise statistics and return updated result

        Args:
            data (dict[str, pd.DataFrame]): The data to be evaluated.
            col (str): Column name
            info (dict[str, Any]): Column information
            method (str): Statistical method name

        Returns:
            (dict[str, float]) Updated column result
        """
        infer_type: list[str] = self.INFER_DTYPE_MAP[method]
        module: callable = self.MODULE_MAP[method]

        result: dict[str, float] = {}
        if info["infer_dtype_match"] and info["ori_infer_dtype"] in infer_type:
            self._logger.debug(
                f"Column '{col}' data type matches required type for method '{method}'"
            )
            for source in ["ori", "syn"]:
                result[f"{method}_{source}"] = module().eval(
                    data={"col": data[source][col]}
                )
        else:
            self._logger.debug(
                f"Column '{col}' data type does not match required type for method '{method}', returning NaN"
            )
            result[f"{method}_ori"] = np.nan
            result[f"{method}_syn"] = np.nan

        return result

    def _process_percolumn(
        self, data: dict[str, pd.DataFrame], col: str, info: dict[str, Any], method: str
    ) -> dict[str, float]:
        """
        Process per-column statistics and return updated result

        Args:
            data (dict[str, pd.DataFrame]): The data to be evaluated.
            col (str): Column name
            info (dict[str, Any]): Column information
            method (str): Statistical method name

        Returns:
            (dict[str, float]) Updated column result
        """
        self._logger.debug(
            f"Processing per-column statistics for column '{col}' using method '{method}'"
        )

        infer_type: list[str] = self.INFER_DTYPE_MAP[method]
        module: callable = self.MODULE_MAP[method]

        result: dict[str, float] = {}
        if info["infer_dtype_match"] and info["ori_infer_dtype"] in infer_type:
            result[method] = module().eval(
                {
                    "col_ori": data["ori"][col],
                    "col_syn": data["syn"][col],
                }
            )
        else:
            result[method] = np.nan

        return result

    def process_pairwise(
        self,
        data: dict[str, pd.DataFrame],
        col1: str,
        info1: dict[str, Any],
        col2: str,
        info2: dict[str, Any],
        method: str,
    ) -> dict[str, float]:
        """
        Process pairwise statistics and return updated result

        Args:
            data (dict[str, pd.DataFrame]): The data to be evaluated.
            col1 (str): First column name
            info1 (dict[str, Any]: First column information
            col2 (str): Second column name
            info2 (dict[str, Any]): Second column information
            method (str): Statistical method name

        Returns:
            (dict[str, float]): Updated pair result
        """
        infer_type = self.INFER_DTYPE_MAP[method]
        module = self.MODULE_MAP[method]

        result = {}
        # Process original data
        if (
            info1["ori_infer_dtype"] in infer_type
            and info2["ori_infer_dtype"] in infer_type
        ):
            result[f"{method}_ori"] = module().eval(
                {
                    "col1": data["ori"][col1],
                    "col2": data["ori"][col2],
                }
            )
        else:
            result[f"{method}_ori"] = np.nan

        # Process synthetic data
        if (
            info1["syn_infer_dtype"] in infer_type
            and info2["syn_infer_dtype"] in infer_type
        ):
            result[f"{method}_syn"] = module().eval(
                {
                    "col1": data["syn"][col1],
                    "col2": data["syn"][col2],
                }
            )
        else:
            result[f"{method}_syn"] = np.nan

        return result

    def _apply_comparison(self, df: pd.DataFrame, compare_method: str) -> pd.DataFrame:
        """
        Apply comparison method to DataFrame

        Args:
            df (pd.DataFrame): The DataFrame to be compared
            compare_method (str): The comparison method
        """
        self._logger.debug(
            f"Applying comparison method '{compare_method}' to DataFrame"
        )
        error_msg: Optional[list[str]] = None

        method_info = self.COMPARE_METHOD_MAP.get(compare_method)
        if not method_info:
            error_msg = f"Unsupported comparison method: {compare_method}"
            self._logger.error(error_msg)
            raise UnsupportedMethodError(error_msg)

        ori_cols: list[str] = [col for col in df.columns if col.endswith("_ori")]
        syn_cols: list[str] = [col.replace("_ori", "_syn") for col in ori_cols]
        self._logger.debug(f"Found {len(ori_cols)} original columns for comparison")

        if not all(col in df.columns for col in syn_cols):
            error_msg = "Missing synthetic columns"
            self._logger.error(error_msg)
            raise ValueError(error_msg)

        result = df.copy()
        for ori_col, syn_col in zip(ori_cols, syn_cols):
            eval_col = f"{ori_col.replace('_ori', f'_{compare_method}')}"

            if eval_col in result.columns:
                result.drop(columns=[eval_col], inplace=True)

            result.insert(result.columns.get_loc(syn_col) + 1, eval_col, np.nan)

            # Apply appropriate function based on whether original value is zero
            func = method_info["func"]
            handle_zero = method_info["handle_zero"]
            result[eval_col] = np.where(
                result[ori_col].astype(float) == 0.0,
                handle_zero(result[syn_col], result[ori_col]),
                safe_round(func(result[syn_col], result[ori_col])),
            )

        self._logger.debug(f"Comparison method '{compare_method}' applied successfully")
        return result

    def _eval(self, data: dict[str, pd.DataFrame]) -> dict[str, pd.DataFrame]:
        """
        Evaluating the evaluator.
            _impl should be initialized in this method.

        Args:
            data (dict[str, pd.DataFrame]): The data to be evaluated.

        Return:
            (dict[str, pd.DataFrame]): The evaluation result
        """
        self._logger.info(
            f"Starting evaluation with {len(self.stats_config.stats_method)} statistical methods"
        )
        self.stats_config.update_data(data)

        columns_results: dict[str, dict[str, float]] = {}
        pairs_results: dict[str, dict[str, float]] = {}
        # Process each column
        for col1, info1 in self.stats_config.columns_info.items():
            columns_results[col1] = {}

            # Process each statistical method
            for method in self.stats_config.stats_method:
                exec_granularity = self.EXEC_GRANULARITY_MAP[method]

                # Update results based on granularity
                if exec_granularity == "columnwise":
                    columns_results[col1].update(
                        self._process_columnwise(data, col1, info1, method)
                    )
                elif exec_granularity == "percolumn":
                    columns_results[col1].update(
                        self._process_percolumn(data, col1, info1, method)
                    )
        self._logger.debug(f"Processed statistics for {len(columns_results)} columns")

        # Process pairwise statistics separately to avoid duplication
        for method in self.stats_config.stats_method:
            if self.EXEC_GRANULARITY_MAP[method] == "pairwise":
                column_items = list(self.stats_config.columns_info.items())
                for i in range(len(column_items)):
                    col1, info1 = column_items[i]
                    for j in range(i + 1, len(column_items)):
                        col2, info2 = column_items[j]
                        pairs_key = (col1, col2)
                        if pairs_key not in pairs_results:
                            pairs_results[pairs_key] = {}
                        pairs_results[pairs_key].update(
                            self._process_pairwise(
                                data, col1, info1, col2, info2, method
                            )
                        )
        if pairs_results:
            self._logger.debug(
                f"Processed pair-wise statistics for {len(pairs_results)} column pairs"
            )

        stats_result: dict[str, pd.DataFrame] = {}
        if columns_results:
            stats_result["columnwise"] = pd.DataFrame(columns_results).T
        if pairs_results:
            stats_result["pairwise"] = pd.DataFrame(pairs_results).T

        # Compare, Aggregated, and Summary results
        compare_method: str = self.stats_config.compare_method
        aggregated_method: str = self.stats_config.aggregated_method
        summary_method: str = self.stats_config.summary_method

        compare_col: list[str] = []
        global_result: dict[str, float] = {}
        for granularity in ["columnwise", "pairwise"]:
            if granularity in stats_result and stats_result is not None:
                # Apply comparison method
                stats_result[granularity] = self._apply_comparison(
                    stats_result[granularity], compare_method
                )
                self._logger.debug(
                    f"Applied comparison method '{compare_method}' to results"
                )

                compare_col = [
                    col
                    for col in stats_result[granularity]
                    if col.endswith(f"_{compare_method}")
                ]
                # add aggregated percolumn method
                compare_col += [
                    stats_method
                    for stats_method in self.stats_config.stats_method
                    if self.EXEC_GRANULARITY_MAP[stats_method] == "percolumn"
                ]

                # Apply aggregated method
                global_result.update(
                    self.AGGREGATED_METHOD_MAP[aggregated_method](
                        stats_result[granularity][compare_col]
                    )
                )

        # Apply summary method
        score: float = self.SUMMARY_METHOD_MAP[summary_method](global_result.values())
        global_result = {"Score": score, **global_result}
        stats_result["global"] = pd.DataFrame.from_dict(global_result, orient="index").T
        self._logger.info(f"Evaluation completed with global score: {score:.4f}")
        return stats_result



"""Pure functions for type inference operations"""

import re
from typing import Any, Dict, Optional, Tuple

import numpy as np
import pandas as pd
from pandas.api.types import (
    is_float_dtype,
    is_integer_dtype,
    is_numeric_dtype,
    is_object_dtype,
)

from petsard.metadater.types.data_types import DataType, LogicalType


def infer_pandas_dtype(series: pd.Series) -> str:
    """
    Pure function to infer pandas dtype from series

    Args:
        series: Pandas series to analyze

    Returns:
        String representation of pandas dtype
    """
    if isinstance(series.dtype, np.dtype):
        return series.dtype.name
    elif isinstance(series.dtype, pd.CategoricalDtype):
        return f"category[{series.dtype.categories.dtype.name}]"
    elif isinstance(series.dtype, pd.api.extensions.ExtensionDtype):
        return str(series.dtype)
    else:
        return str(series.dtype)


def map_pandas_to_metadater_type(pandas_dtype: str) -> DataType:
    """
    Pure function to convert pandas dtype to Metadater DataType

    Args:
        pandas_dtype: Pandas dtype string

    Returns:
        Corresponding DataType enum
    """
    dtype_str = pandas_dtype.lower()

    # Basic type mapping
    type_mapping = {
        "bool": DataType.BOOLEAN,
        "int8": DataType.INT8,
        "int16": DataType.INT16,
        "int32": DataType.INT32,
        "int64": DataType.INT64,
        "uint8": DataType.INT16,  # Promote unsigned to next larger signed
        "uint16": DataType.INT32,
        "uint32": DataType.INT64,
        "uint64": DataType.INT64,
        "float16": DataType.FLOAT32,
        "float32": DataType.FLOAT32,
        "float64": DataType.FLOAT64,
        "object": DataType.STRING,
        "string": DataType.STRING,
        "category": DataType.STRING,
    }

    # Handle datetime types
    if dtype_str.startswith("datetime64"):
        if "utc" in dtype_str.lower() or "tz" in dtype_str.lower():
            return DataType.TIMESTAMP_TZ
        return DataType.TIMESTAMP

    # Handle category types
    if dtype_str.startswith("category"):
        return DataType.STRING

    return type_mapping.get(dtype_str, DataType.STRING)


def detect_logical_type_patterns(
    series: pd.Series,
    sample_size: Optional[int] = 1000,
    confidence_threshold: float = 0.95,
) -> Optional[LogicalType]:
    """
    Pure function to detect logical type from data patterns

    Args:
        series: Pandas series to analyze
        sample_size: Maximum number of samples to analyze
        confidence_threshold: Minimum confidence required for detection

    Returns:
        Detected LogicalType or None
    """
    # Only analyze string-like data
    if not _is_string_compatible(series):
        return None

    sample = series.dropna()
    if len(sample) == 0:
        return None

    # Limit sample size for performance
    if sample_size and len(sample) > sample_size:
        sample = sample.sample(n=sample_size, random_state=42)

    # Ensure string operations are possible
    try:
        if not pd.api.types.is_string_dtype(sample):
            sample = sample.astype(str)
    except Exception:
        return None

    # Get pattern definitions
    patterns = _get_logical_type_patterns()

    # Check each pattern
    for logical_type, (pattern, threshold) in patterns.items():
        try:
            matches = sample.str.match(pattern, na=False)
            match_ratio = matches.sum() / len(matches) if len(matches) > 0 else 0.0

            if match_ratio >= max(threshold, confidence_threshold):
                # Special validation for certain types
                if logical_type in [
                    LogicalType.LATITUDE,
                    LogicalType.LONGITUDE,
                    LogicalType.PERCENTAGE,
                    LogicalType.CURRENCY,
                ]:
                    if not _validate_special_logical_type(
                        sample, logical_type, threshold
                    ):
                        continue
                return logical_type
        except Exception:
            continue

    # Check for categorical data (low cardinality)
    unique_ratio = series.nunique() / len(series)
    if unique_ratio < 0.05:  # 5% threshold
        return LogicalType.CATEGORICAL

    return None


def infer_optimal_dtype(
    series: pd.Series,
    current_dtype: Optional[str] = None,
    logical_type: Optional[LogicalType] = None,
) -> str:
    """
    Pure function to infer optimal dtype for storage

    Args:
        series: Pandas series to analyze
        current_dtype: Current dtype string
        logical_type: Detected logical type

    Returns:
        Optimal dtype string
    """
    # 首先檢查前導零情況
    if _has_leading_zeros_inference(series):
        return "string"

    # If logical type suggests category, use category
    if logical_type == LogicalType.CATEGORICAL:
        return "category"

    # For numeric data, find the smallest suitable type
    if is_numeric_dtype(series):
        return _optimize_numeric_dtype(series)

    # For object data, try to optimize
    elif is_object_dtype(series):
        return _optimize_object_dtype(series)

    # For other types, keep current
    return current_dtype or str(series.dtype)


def detect_datetime_format(series: pd.Series, sample_size: int = 100) -> Optional[str]:
    """
    Pure function to detect datetime format from string data

    Args:
        series: Series with potential datetime strings
        sample_size: Number of samples to analyze

    Returns:
        Detected datetime format string or None
    """
    if not _is_string_compatible(series):
        return None

    sample = series.dropna().head(sample_size)
    if len(sample) == 0:
        return None

    # Common datetime formats to try
    formats = [
        "%Y-%m-%d",
        "%Y/%m/%d",
        "%d-%m-%Y",
        "%d/%m/%Y",
        "%Y-%m-%d %H:%M:%S",
        "%Y/%m/%d %H:%M:%S",
        "%d-%m-%Y %H:%M:%S",
        "%d/%m/%Y %H:%M:%S",
        "%Y-%m-%dT%H:%M:%S",
        "%Y-%m-%dT%H:%M:%SZ",
        "%Y-%m-%d %H:%M:%S.%f",
        "%m/%d/%Y",
        "%m-%d-%Y",
    ]

    for fmt in formats:
        try:
            # Try to parse a few samples
            parsed_count = 0
            for value in sample.head(10):
                try:
                    pd.to_datetime(str(value), format=fmt)
                    parsed_count += 1
                except Exception:
                    continue

            # If most samples parse successfully, this is likely the format
            if parsed_count >= len(sample.head(10)) * 0.8:
                return fmt
        except Exception:
            continue

    return None


def analyze_cardinality(series: pd.Series) -> Dict[str, Any]:
    """
    Pure function to analyze cardinality characteristics

    Args:
        series: Series to analyze

    Returns:
        Dictionary with cardinality analysis
    """
    total_count = len(series)
    unique_count = series.nunique()
    null_count = series.isna().sum()

    if total_count == 0:
        return {
            "total_count": 0,
            "unique_count": 0,
            "null_count": 0,
            "cardinality_ratio": 0.0,
            "is_unique": False,
            "is_low_cardinality": False,
            "is_high_cardinality": False,
        }

    cardinality_ratio = unique_count / total_count

    return {
        "total_count": int(total_count),
        "unique_count": int(unique_count),
        "null_count": int(null_count),
        "cardinality_ratio": round(cardinality_ratio, 4),
        "is_unique": cardinality_ratio >= 0.95,
        "is_low_cardinality": cardinality_ratio <= 0.05,
        "is_high_cardinality": cardinality_ratio >= 0.8,
    }


# Helper functions


def _is_string_compatible(series: pd.Series) -> bool:
    """Check if series can use string operations"""
    if pd.api.types.is_string_dtype(series):
        return True
    if pd.api.types.is_object_dtype(series):
        try:
            non_null = series.dropna()
            if len(non_null) == 0:
                return False
            return all(isinstance(x, str) for x in non_null.head(100))
        except Exception:
            return False
    return False


def _get_logical_type_patterns() -> Dict[LogicalType, Tuple[str, float]]:
    """Get logical type detection patterns with confidence thresholds"""
    return {
        LogicalType.EMAIL: (r"^[\w\.-]+@[\w\.-]+\.\w+$", 0.95),
        LogicalType.URL: (r"^https?://[^\s]+$", 0.95),
        LogicalType.IP_ADDRESS: (
            r"^(?:(?:25[0-5]|2[0-4][0-9]|[01]?[0-9][0-9]?)\.){3}(?:25[0-5]|2[0-4][0-9]|[01]?[0-9][0-9]?)$",
            0.95,
        ),
        LogicalType.UUID: (
            r"^[0-9a-fA-F]{8}-[0-9a-fA-F]{4}-[0-9a-fA-F]{4}-[0-9a-fA-F]{4}-[0-9a-fA-F]{12}$",
            0.98,
        ),
        LogicalType.PHONE: (r"^[\+]?[\d\s\-\(\)]{10,}$", 0.90),
        LogicalType.POSTAL_CODE: (r"^[\d\w\s\-]{3,10}$", 0.85),
    }


def _validate_special_logical_type(
    sample: pd.Series, logical_type: LogicalType, threshold: float
) -> bool:
    """Validate special logical types with custom logic"""
    validators = {
        LogicalType.LATITUDE: _validate_latitude_values,
        LogicalType.LONGITUDE: _validate_longitude_values,
        LogicalType.PERCENTAGE: _validate_percentage_values,
        LogicalType.CURRENCY: _validate_currency_values,
    }

    validator = validators.get(logical_type)
    if not validator:
        return True

    try:
        valid_count = sum(validator(str(x)) for x in sample)
        return (valid_count / len(sample)) >= threshold
    except Exception:
        return False


def _validate_latitude_values(value: str) -> bool:
    """Validate latitude value"""
    try:
        lat = float(value)
        return -90 <= lat <= 90
    except (ValueError, TypeError):
        return False


def _validate_longitude_values(value: str) -> bool:
    """Validate longitude value"""
    try:
        lon = float(value)
        return -180 <= lon <= 180
    except (ValueError, TypeError):
        return False


def _validate_percentage_values(value: str) -> bool:
    """Validate percentage value"""
    if value.endswith("%"):
        try:
            float(value[:-1])
            return True
        except ValueError:
            return False
    try:
        num = float(value)
        return 0 <= num <= 1 or 0 <= num <= 100
    except ValueError:
        return False


def _validate_currency_values(value: str) -> bool:
    """Validate currency value"""
    currency_patterns = [
        r"^\$[\d,]+\.?\d*$",
        r"^[\d,]+\.?\d*\$$",
        r"^€[\d,]+\.?\d*$",
        r"^£[\d,]+\.?\d*$",
        r"^¥[\d,]+\.?\d*$",
        r"^[A-Z]{3}\s?[\d,]+\.?\d*$",
    ]
    return any(re.match(pattern, value) for pattern in currency_patterns)


def _is_float_with_integer_values_inference(series: pd.Series) -> bool:
    """
    檢測 float 序列是否實際上都是整數值（type_inference 版本）

    Args:
        series: 要檢查的 Series

    Returns:
        True 如果所有非空值都是整數，False 否則
    """
    if not is_float_dtype(series):
        return False

    # 移除空值
    clean_data = series.dropna()
    if len(clean_data) == 0:
        return False

    # 檢查所有值是否都是整數
    try:
        return all(val.is_integer() for val in clean_data)
    except (AttributeError, TypeError):
        return False


def _optimize_numeric_dtype(series: pd.Series) -> str:
    """Optimize numeric dtype to smallest suitable type"""
    if is_integer_dtype(series):
        # 如果有空值，使用 nullable integer 避免轉為 float
        if series.isna().any():
            if series.isna().all():
                return "Int64"

            col_min, col_max = np.nanmin(series), np.nanmax(series)

            # 使用 nullable integer types
            ranges = [
                ("Int8", np.iinfo(np.int8).min, np.iinfo(np.int8).max),
                ("Int16", np.iinfo(np.int16).min, np.iinfo(np.int16).max),
                ("Int32", np.iinfo(np.int32).min, np.iinfo(np.int32).max),
                ("Int64", np.iinfo(np.int64).min, np.iinfo(np.int64).max),
            ]

            for dtype, min_val, max_val in ranges:
                if min_val <= col_min and col_max <= max_val:
                    return dtype
            return "Int64"
        else:
            # 沒有空值時使用傳統 integer types
            col_min, col_max = np.nanmin(series), np.nanmax(series)

            ranges = [
                ("int8", np.iinfo(np.int8).min, np.iinfo(np.int8).max),
                ("int16", np.iinfo(np.int16).min, np.iinfo(np.int16).max),
                ("int32", np.iinfo(np.int32).min, np.iinfo(np.int32).max),
                ("int64", np.iinfo(np.int64).min, np.iinfo(np.int64).max),
            ]

            for dtype, min_val, max_val in ranges:
                if min_val <= col_min and col_max <= max_val:
                    return dtype
            return "int64"

    elif is_float_dtype(series):
        # 檢查是否實際上是整數值
        if _is_float_with_integer_values_inference(series):
            # 轉換為 nullable integer
            if series.isna().all():
                return "Int64"

            col_min, col_max = np.nanmin(series), np.nanmax(series)

            ranges = [
                ("Int8", np.iinfo(np.int8).min, np.iinfo(np.int8).max),
                ("Int16", np.iinfo(np.int16).min, np.iinfo(np.int16).max),
                ("Int32", np.iinfo(np.int32).min, np.iinfo(np.int32).max),
                ("Int64", np.iinfo(np.int64).min, np.iinfo(np.int64).max),
            ]

            for dtype, min_val, max_val in ranges:
                if min_val <= col_min and col_max <= max_val:
                    return dtype
            return "Int64"
        else:
            # 真正的浮點數
            if series.isna().all():
                return "float32"

            col_min, col_max = np.nanmin(series), np.nanmax(series)

            # Check if float32 is sufficient
            if (
                np.finfo(np.float32).min <= col_min
                and col_max <= np.finfo(np.float32).max
            ):
                return "float32"
            else:
                return "float64"

    return str(series.dtype)


def _has_leading_zeros_inference(series: pd.Series) -> bool:
    """
    檢測字串是否包含前導零的數字（type_inference 版本）

    Args:
        series: 要檢查的 Series

    Returns:
        True 如果發現前導零模式，False 否則
    """
    if series.isna().all():
        return False

    clean_series = series.dropna()
    if len(clean_series) == 0:
        return False

    # 轉換為字串進行檢查
    try:
        series_str = clean_series.astype(str)
    except Exception:
        return False

    # 檢查前導零模式：開頭是 "0" 且後面跟數字
    leading_zero_count = 0
    total_checked = 0

    for value in series_str.head(100):  # 檢查前100個值以提高效能
        value = value.strip()
        if len(value) >= 2:  # 至少要有兩個字符
            # 檢查是否符合前導零模式：0開頭且全部都是數字
            if value.startswith("0") and value.isdigit():
                leading_zero_count += 1
            total_checked += 1

    # 如果超過30%的值有前導零，則認為這是前導零欄位
    if total_checked > 0:
        return (leading_zero_count / total_checked) >= 0.3

    return False


def _optimize_object_dtype(series: pd.Series) -> str:
    """Optimize object dtype"""
    if series.isna().all():
        return "category"

    clean_series = series.dropna()

    # 首先檢查是否有前導零，如果有則保持為字串
    if _has_leading_zeros_inference(series):
        return "string"

    # Try datetime conversion
    try:
        datetime_series = pd.to_datetime(clean_series, errors="coerce")
        if not datetime_series.isna().any():
            return "datetime64[ns]"
    except Exception:
        pass

    # Default to category for object types
    return "category"



import numpy as np
import pandas as pd
import pytest

from petsard import Constrainer


class MockSynthesizer:
    def sample(self, num_rows: int) -> pd.DataFrame:
        return pd.DataFrame(
            {
                "name": np.random.choice(
                    ["John", "Mary", "Tom", "Jane", None], num_rows
                ),
                "job": np.random.choice(
                    ["Engineer", "Doctor", "Teacher", None], num_rows
                ),
                "salary": np.random.randint(30000, 120000, num_rows),
                "bonus": [
                    None if np.random.random() < 0.3 else x / 10
                    for x in np.random.randint(30000, 120000, num_rows)
                ],
                "age": np.random.randint(20, 70, num_rows),
                "education": np.random.choice(
                    ["High School", "Bachelor", "Master", "PhD"], num_rows
                ),
                "performance": np.random.randint(1, 6, num_rows),
            }
        )


@pytest.fixture
def sample_df():
    return pd.DataFrame(
        {
            "name": ["John", None, "Mary", "Tom", "Jane"],
            "job": ["Engineer", "Doctor", None, "Teacher", None],
            "salary": [50000, 80000, None, 60000, 70000],
            "bonus": [5000, None, 8000, None, 7000],
            "age": [25, 35, 45, 55, 65],
            "education": ["High School", "Bachelor", "Master", "PhD", "PhD"],
            "performance": [3, 4, 5, 4, 5],
        }
    )


@pytest.fixture
def config():
    return {
        "nan_groups": {
            "name": {"delete": "salary"},
            "job": {"erase": ["salary", "bonus"]},
            "salary": {"copy": "bonus"},
        },
        "field_constraints": ["age >= 20 & age <= 60", "performance >= 4"],
        "field_combinations": [
            (
                {"education": "performance"},
                {"PhD": [4, 5], "Master": [4, 5], "Bachelor": [3, 4, 5]},
            )
        ],
    }


def test_basic_initialization(config):
    """Test basic constrainer initialization"""
    constrainer = Constrainer(config)
    assert constrainer is not None
    assert constrainer.config == config


def test_nan_groups_constraints(sample_df, config):
    """Test NaN group constraints application"""
    constrainer = Constrainer({"nan_groups": config["nan_groups"]})
    result = constrainer.apply(sample_df)

    # Test 'delete' action
    assert all(pd.notna(result["name"]))

    # Test 'erase' action
    job_null_mask = pd.isna(result["job"])
    assert all(pd.isna(result.loc[job_null_mask, "salary"]))
    assert all(pd.isna(result.loc[job_null_mask, "bonus"]))

    # Test 'copy' action
    salary_mask = pd.notna(result["salary"]) & pd.isna(result["bonus"])
    if not result[salary_mask].empty:
        assert all(
            result.loc[salary_mask, "salary"] == result.loc[salary_mask, "bonus"]
        )


def test_field_constraints(sample_df, config):
    """Test field constraints application"""
    constrainer = Constrainer({"field_constraints": config["field_constraints"]})
    result = constrainer.apply(sample_df)

    # Test age constraint
    assert all(result["age"].between(20, 60))

    # Test performance constraint
    assert all(result["performance"] >= 4)


def test_field_combinations(sample_df, config):
    """Test field combinations constraints"""
    constrainer = Constrainer({"field_combinations": config["field_combinations"]})
    result = constrainer.apply(sample_df)

    # Test education-performance combinations
    phd_mask = result["education"] == "PhD"
    assert all(result.loc[phd_mask, "performance"].isin([4, 5]))

    master_mask = result["education"] == "Master"
    assert all(result.loc[master_mask, "performance"].isin([4, 5]))


def test_all_constraints_together(sample_df, config):
    """Test all constraints working together"""
    constrainer = Constrainer(config)
    result = constrainer.apply(sample_df)

    # Should meet all conditions
    assert all(pd.notna(result["name"]))
    assert all(result["age"].between(20, 60))
    assert all(result["performance"] >= 4)

    phd_mask = result["education"] == "PhD"
    if not result[phd_mask].empty:
        assert all(result.loc[phd_mask, "performance"].isin([4, 5]))


def test_resample_functionality(sample_df, config):
    """Test resample_until_satisfy functionality"""
    constrainer = Constrainer(config)
    synthesizer = MockSynthesizer()

    target_rows = 5
    result = constrainer.resample_until_satisfy(
        data=sample_df,
        target_rows=target_rows,
        synthesizer=synthesizer,
        sampling_ratio=2.0,
        max_trials=10,
    )

    # Check basic requirements
    assert len(result) == target_rows
    assert all(result["age"].between(20, 60))
    assert all(result["performance"] >= 4)


def test_error_handling(sample_df):
    """Test error handling"""
    # Test invalid config format
    with pytest.raises(ValueError):
        Constrainer("not a dict")

    # Test missing columns
    invalid_config = {"field_constraints": ["invalid_column > 0"]}
    constrainer = Constrainer(invalid_config)
    with pytest.raises(Exception):
        constrainer.apply(sample_df)


def test_edge_cases(sample_df, config):
    """Test edge cases"""
    constrainer = Constrainer(config)

    # Test empty DataFrame
    empty_df = pd.DataFrame(columns=sample_df.columns)
    result = constrainer.apply(empty_df)
    assert result.empty

    # Test DataFrame with all NaN
    all_nan_df = pd.DataFrame(
        {
            "name": [None] * 3,
            "job": [None] * 3,
            "salary": [None] * 3,
            "bonus": [None] * 3,
            "age": [None] * 3,
            "education": [None] * 3,
            "performance": [None] * 3,
        }
    )
    result = constrainer.apply(all_nan_df)
    assert result.empty
